import json
import logging
from timeit import default_timer
import warnings
import datetime as dt
from hashlib import md5
from time import sleep
from typing import Any
from urllib import parse as urlparse

from bravado import requests_client
from bravado.client import SwaggerClient
from bravado.exception import (
    HTTPBadGateway, HTTPGatewayTimeout, HTTPServiceUnavailable, HTTPError,
)
from bravado.http_future import HttpFuture
from bravado.swagger_model import Loader
from bravado_core.response import IncomingResponse
from bravado_core.spec import CONFIG_DEFAULTS, Spec
from requests.adapters import HTTPAdapter

from django.core.cache import cache

from . import __title__, __url__, __version__, app_settings
from .errors import TokenExpiredError
from .signals import esi_request_statistics

logger = logging.getLogger(__name__)

_LIBRARIES_LOG_LEVEL = logging.getLevelName(app_settings.ESI_LOG_LEVEL_LIBRARIES)
logging.getLogger('swagger_spec_validator').setLevel(_LIBRARIES_LOG_LEVEL)
logging.getLogger('bravado_core').setLevel(_LIBRARIES_LOG_LEVEL)
logging.getLogger('urllib3').setLevel(_LIBRARIES_LOG_LEVEL)
logging.getLogger('bravado').setLevel(_LIBRARIES_LOG_LEVEL)

SPEC_CONFIG = {'use_models': False}
RETRY_SLEEP_SECS = 1


class CachingHttpFuture(HttpFuture):
    """Extended wrapper for a FutureAdapter that returns a HTTP response
    and also supports caching.

    This class contains the response for an ESI request with an ESI client.
    """
    def _cache_key(self) -> str:
        """Generate the key name used to cache responses."""
        request = self.future.request
        data = (
            request.method
            + request.url
            + str(request.params)
            + str(request.data)
            + str(request.json)
        ).encode('utf-8')
        # The following hash is not used in any security context. It is only used
        # to generate unique values, collisions are acceptable and "data" is not
        # coming from user-generated input
        str_hash = md5(data).hexdigest()  # nosec B303, B303-1
        return f'esi_{str_hash}'

    @staticmethod
    def _time_to_expiry(expires):
        """Determine the seconds until a HTTP header "Expires" timestamp.

        Args:
            expires: HTTP response "Expires" header

        Returns:
            seconds until "Expires" time
        """
        try:
            expires_dt = dt.datetime.strptime(str(expires), '%a, %d %b %Y %H:%M:%S %Z')
            if expires_dt.tzinfo is None:
                expires_dt = expires_dt.replace(tzinfo=dt.timezone.utc)
            delta = expires_dt - dt.datetime.now(dt.timezone.utc)
            return delta.total_seconds()
        except ValueError:
            return 0

    def results(self, **kwargs) -> Any | tuple[Any, IncomingResponse]:
        """Executes the request and returns the response from ESI for the current
        route. Response will include all pages if there are more available.

        Accepts same parameters in ``kwargs`` as :meth:`result`

        Returns:
            same as :meth:`result`, but for multiple pages
        """
        results = list()
        headers = None
        # preserve original value
        _also_return_response = self.request_config.also_return_response
        # override to always get the raw response for expiry header
        self.request_config.also_return_response = True

        if "page" in self.operation.params:
            current_page = 1
            total_pages = 1

            # loop all pages and add data to output array
            while current_page <= total_pages:
                self.future.request.params["page"] = current_page
                # will use cache if applicable
                result, headers = self.result(**kwargs)
                total_pages = int(headers.headers['X-Pages'])
                # append to results list to be seamless to the client
                results += result
                current_page += 1
        else:  # it doesn't so just return
            results, headers = self.result(**kwargs)

        # restore original value
        self.request_config.also_return_response = _also_return_response

        # obey the output
        if self.request_config.also_return_response:
            return results, headers
        else:
            return results

    def results_localized(self, languages: list = None, **kwargs) -> dict:
        """Executes the request and returns the response from ESI for all default
        languages and pages (if any).

        Accepts same parameters in ``kwargs`` as :meth:`result` plus ``languages``

        Args:
            languages: (optional) list of languages to return \
                instead of default languages

        Returns:
            Dict of all responses with the language code as keys.
        """
        if not languages:
            my_languages = list(app_settings.ESI_LANGUAGES)
        else:
            my_languages = []
            for lang in dict.fromkeys(languages):
                if lang not in app_settings.ESI_LANGUAGES:
                    raise ValueError('Invalid language code: %s' % lang)
                my_languages.append(lang)

        return {
            language: self.results(language=language, **kwargs)
            for language in my_languages
        }

    def _send_signal(self, status_code: int, headers: dict = {}, latency: float = 0) -> None:
        """
            Dispatch the esi request statistics signal
        """
        esi_request_statistics.send(
            sender=self.__class__,
            operation=self.operation.path_name,
            status_code=status_code,
            headers=headers,
            latency=latency,
            bucket=""
        )

    def result(self, **kwargs) -> Any | tuple[Any, IncomingResponse]:
        """Executes the request and returns the response from ESI. Response will
        include the requested / first page only if there are more pages available.

        Args:
            timeout: (optional) timeout for ESI request in seconds, overwrites default
            retries: (optional) max number of retries, overwrites default
            language: (optional) retrieve result for specific language
            ignore_cache: (optional) set to ``True`` to ignore response caching

        Returns:
            Response from endpoint or a tuple with response from endpoint \
            and an incoming response object containing additional meta data \
            including the HTTP response headers
        """
        if 'language' in kwargs.keys():
            # this parameter is not supported by bravado, so we can't pass it on
            self.future.request.params['language'] = str(kwargs.pop('language'))

        if 'timeout' not in kwargs:
            kwargs['timeout'] = (
                app_settings.ESI_REQUESTS_CONNECT_TIMEOUT,
                app_settings.ESI_REQUESTS_READ_TIMEOUT
            )

        ignore_cache = (
            kwargs.pop('ignore_cache') if 'ignore_cache' in kwargs.keys() else False
        )

        if (
            app_settings.ESI_CACHE_RESPONSE
            and not ignore_cache
            and self.future.request.method == 'GET'
            and self.operation is not None
        ):
            result = None
            response = None
            cache_key = self._cache_key()
            try:
                cached = cache.get(cache_key)
            except Exception:
                cached = None
                logger.warning(
                    "Attempt to read ESI results from cache failed", exc_info=True
                )

            if cached:
                self._send_signal(
                    status_code=0
                )
                result, response = cached
                expiry = self._time_to_expiry(str(response.headers.get('Expires')))
                if expiry < 0:
                    logger.warning(
                        "cache expired by %d seconds, Forcing expiry", expiry
                    )
                    cached = False

            if not cached:
                result, response = self._result_with_retries(**kwargs)
                if response and 'Expires' in response.headers:
                    expires = self._time_to_expiry(response.headers['Expires'])
                    if expires > 0:
                        try:
                            cache.set(cache_key, (result, response), expires)
                        except Exception:
                            logger.warning(
                                "Failed to write ESI result to cache", exc_info=True
                            )

            if self.request_config.also_return_response:
                return result, response
            return result

        elif self.operation is not None:
            result, response = self._result_with_retries(**kwargs)
            if self.request_config.also_return_response:
                return result, response
            return result

        return super().result(**kwargs)

    def _result_with_retries(self, **kwargs) -> tuple[Any, IncomingResponse]:
        """Execute request and retry on certain HTTP errors.

        ``kwargs`` are passed through to super().result()

        Returns:
            Tuple with response from endpoint and an incoming response object \
                containing additional meta data including the HTTP response headers
        """
        # preserve original value
        _also_return_response = self.request_config.also_return_response
        # override to always get the raw response for expiry header
        self.request_config.also_return_response = True

        if 'retries' in kwargs.keys():
            max_retries = int(kwargs.pop('retries'))
        else:
            max_retries = int(app_settings.ESI_SERVER_ERROR_MAX_RETRIES)
        max_retries = max(0, max_retries)

        retries = 0
        while retries <= max_retries:
            _t = default_timer()
            try:
                if app_settings.ESI_INFO_LOGGING_ENABLED:
                    params = self.future.request.params
                    logger.info(
                        'Fetching from ESI: %s%s%s',
                        self.future.request.url,
                        f' - language {params["language"]}'
                        if 'language' in params else '',
                        f' - page {params["page"]}'
                        if 'page' in params else ''
                    )
                logger.debug(
                    'ESI request: %s - %s',
                    self.future.request.url,
                    self.future.request.params
                )
                logger.debug('ESI request headers: %s', self.future.request.headers)
                result, response = super().result(**kwargs)
                logger.debug('ESI response status code: %s', response.status_code)
                logger.debug('ESI response headers: %s', response.headers)
                if app_settings.ESI_DEBUG_RESPONSE_CONTENT_LOGGING:
                    logger.debug('ESI response content: %s', response.text)
                break
            except (HTTPBadGateway, HTTPGatewayTimeout, HTTPServiceUnavailable) as ex:
                self._send_signal(
                    status_code=ex.status_code,
                    headers=ex.response.headers,
                    latency=default_timer() - _t
                )
                if retries < max_retries:
                    retries += 1
                    logger.warning(
                        "ESI error - %s %s - Retry: %d/%d",
                        self.future.request.url,
                        ex.status_code,
                        retries,
                        max_retries
                    )
                    wait_secs = (
                        app_settings.ESI_SERVER_ERROR_BACKOFF_FACTOR
                        * (2 ** (retries - 1))
                    )
                    sleep(wait_secs)
                else:
                    raise ex
            except HTTPError as ex:
                """
                    Throw any other error into the signal
                    then just re-raise
                """
                self._send_signal(
                    status_code=ex.status_code,
                    headers=ex.response.headers,
                    latency=default_timer() - _t
                )
                raise ex

        self._send_signal(
            status_code=response.status_code,
            headers=response.headers,
            latency=default_timer() - _t
        )
        # restore original value
        self.request_config.also_return_response = _also_return_response
        return result, response


requests_client.HttpFuture = CachingHttpFuture


class TokenAuthenticator(requests_client.Authenticator):
    """
    Adds the authorization header containing access token, if specified.
    Sets ESI datasource to tranquility or singularity.
    """

    def __init__(self, token=None, datasource=None):
        host = urlparse.urlsplit(app_settings.ESI_API_URL).hostname
        super().__init__(host)
        self.token = token
        self.datasource = datasource

    def apply(self, request):
        if self.token and self.token.expired:
            if self.token.can_refresh:
                self.token.refresh()
            else:
                raise TokenExpiredError()
        request.headers['Authorization'] = \
            'Bearer ' + self.token.access_token if self.token else None
        request.params['datasource'] = \
            self.datasource or app_settings.ESI_API_DATASOURCE
        return request


class RequestsClientPlus(requests_client.RequestsClient):
    """RequestsClient with ability to set the user agent header for all requests"""

    def __init__(
        self,
        ssl_verify=True,
        ssl_cert=None,
        future_adapter_class=requests_client.RequestsFutureAdapter,
        response_adapter_class=requests_client.RequestsResponseAdapter,
    ):
        super().__init__(
            ssl_verify, ssl_cert, future_adapter_class, response_adapter_class
        )
        self.user_agent = None

    def request(
        self, request_params, operation=None, request_config=None
    ) -> HttpFuture:
        if self.user_agent:
            current_headers = request_params.get("headers", dict())
            new_header = {"User-Agent": str(self.user_agent)}
            request_params["headers"] = {**current_headers, **new_header}

        return super().request(request_params, operation, request_config)


def build_cache_name(name):
    """
    Cache key name formatter
    :param name: Name of the spec dict to cache, usually version
    :return: String name for cache key
    :rtype: str
    """
    return 'esi_swaggerspec_%s' % name


def cache_spec(name, spec):
    """
    Cache the spec dict
    :param name: Version name
    :param spec: Spec dict
    :return: True if cached
    """
    return cache.set(
        build_cache_name(name), spec, app_settings.ESI_SPEC_CACHE_DURATION
    )


def build_spec_url(spec_version):
    """
    Generates the URL to swagger.json for the ESI version
    :param spec_version: Name of the swagger spec version, like latest or v4
    :return: URL to swagger.json for the requested spec version
    """
    return urlparse.urljoin(app_settings.ESI_API_URL, spec_version + '/swagger.json')


def get_spec(name, http_client=None, config=None):
    """
    :param name: Name of the revision of spec, eg latest or v4
    :param http_client: Requests client used for retrieving specs
    :param config: Spec configuration - see Spec.CONFIG_DEFAULTS
    :return: :class:`bravado_core.spec.Spec`
    """
    http_client = http_client or requests_client.RequestsClient()

    def load_spec():
        loader = Loader(http_client)
        return loader.load_spec(build_spec_url(name))

    spec_dict = cache.get_or_set(
        build_cache_name(name), load_spec, app_settings.ESI_SPEC_CACHE_DURATION
    )
    config = dict(CONFIG_DEFAULTS, **(config or {}))
    return Spec.from_dict(spec_dict, build_spec_url(name), http_client, config)


def build_spec(base_version, http_client=None, **kwargs):
    """
    Generates the Spec used to initialize a SwaggerClient,
    supporting mixed resource versions
    :param http_client: :class:`bravado.requests_client.RequestsClient`
    :param base_version: Version to base the spec on.
    Any resource without an explicit version will be this.
    :param kwargs: Explicit resource versions, by name (eg Character='v4')
    :return: :class:`bravado_core.spec.Spec`
    """
    base_spec = get_spec(base_version, http_client=http_client, config=SPEC_CONFIG)
    if kwargs:
        for resource, resource_version in kwargs.items():
            versioned_spec = get_spec(
                resource_version, http_client=http_client, config=SPEC_CONFIG
            )
            try:
                spec_resource = versioned_spec.resources[resource.capitalize()]
            except KeyError:
                raise AttributeError(
                    'Resource {} not found on API revision {}'.format(
                        resource, resource_version
                    )
                )
            base_spec.resources[resource.capitalize()] = spec_resource
    return base_spec


def read_spec(path, http_client=None):
    """
    Reads in a swagger spec file used to initialize a SwaggerClient
    :param path: String path to local swagger spec file.
    :param http_client: :class:`bravado.requests_client.RequestsClient`
    :return: :class:`bravado_core.spec.Spec`
    """
    with open(path, encoding='utf-8') as f:
        spec_dict = json.loads(f.read())

    return SwaggerClient.from_spec(
        spec_dict, http_client=http_client, config=SPEC_CONFIG
    )


def esi_client_factory(
        token=None, datasource: str = None, spec_file: str = None, version: str = None,
        app_info_text: str = None,  # Deprecate in favour of the following variables
        ua_appname: str = None, ua_version: str = None, ua_url: str = None, **kwargs) -> SwaggerClient:
    """Generate a new ESI client.

    Args:
        token(esi.models.Token): used to access authenticated endpoints.
        datasource: Name of the ESI datasource to access.
        spec_file: Absolute path to a swagger spec file to load.
        version: Base ESI API version. Accepted values are 'legacy', 'latest',
        ua_appname: Name of the App for generating a User-Agent,
        ua_version: Version of the App for generating a User-Agent,
        ua_url: (optional) URL To the Source Code or Documentation for generating a User-Agent,
        kwargs: Explicit resource versions to build, in the form Character='v4'. \
            Same values accepted as version.

    If a spec_file is specified, specific versioning is not available.
    Meaning the version and resource version kwargs are ignored in favour of the
    versions available in the spec_file.

    Returns:
        New ESI client
    """

    if app_info_text is not None:
        warnings.warn(
            "The 'app_info_text' parameter is deprecated and will be removed in a future release. "
            "Use 'ua_appname', 'ua_version', and `ua_url` to dynamically build a User-Agent instead",
            DeprecationWarning,
            stacklevel=2
        )

    if ua_appname is None or ua_version is None:
        warnings.warn(
            "Applications must define their own 'ua_appname' and 'ua_version' to generate a User-Agent",
            DeprecationWarning,
            stacklevel=2
        )

    if app_settings.ESI_INFO_LOGGING_ENABLED:
        logger.info('Generating an ESI client...')

    client = RequestsClientPlus()

    from esi.helpers import pascal_case_string
    sanitized_appname = pascal_case_string(__title__)

    if app_info_text:
        # app_info_text (email@example) Django-ESI/1.2.3 (+https://gitlab.com/allianceauth/django-esi)
        # Deprecated
        user_agent = f"{app_info_text} ({app_settings.ESI_USER_CONTACT_EMAIL}) {sanitized_appname}/{__version__} (+{__url__})"
    elif ua_appname is None or ua_version is None:
        # Django-ESI/1.2.3 () (email@example; +https://gitlab.com/allianceauth/django-esi)
        # Deprecated
        user_agent = f"{sanitized_appname}/{__version__} ({app_settings.ESI_USER_CONTACT_EMAIL}; +{__url__})"
    else:
        # AppName/1.2.3 (email@example.com) Django-ESI/1.2.3 (+https://gitlab.com/allianceauth/django-esi)
        # or AppName/1.2.3 (email@example.com; +https://gitlab.com/) Django-ESI/1.2.3 (+https://gitlab.com/allianceauth/django-esi) (+https://gitlab.com/allianceauth/django-esi)
        # Preferred

        # Enforce PascalCase for `ua_appname` and strip whitespace
        sanitized_ua_appname = pascal_case_string(ua_appname)

        user_agent = f"{sanitized_ua_appname}/{ua_version} ({app_settings.ESI_USER_CONTACT_EMAIL}{f'; +{ua_url})' if ua_url else ')'} {sanitized_appname}/{__version__} (+{__url__})"

    client.user_agent = user_agent

    my_http_adapter = HTTPAdapter(
        pool_maxsize=app_settings.ESI_CONNECTION_POOL_MAXSIZE,
        max_retries=app_settings.ESI_CONNECTION_ERROR_MAX_RETRIES
    )
    client.session.mount('https://', my_http_adapter)

    if token or datasource:
        client.authenticator = TokenAuthenticator(token=token, datasource=datasource)

    api_version = version or app_settings.ESI_API_VERSION

    if spec_file:
        return read_spec(spec_file, http_client=client)
    else:
        spec = build_spec(api_version, http_client=client, **kwargs)
        return SwaggerClient(spec)


def minimize_spec(spec_dict, operations=None, resources=None):
    """
    Trims down a source spec dict to only the operations or resources indicated.
    :param spec_dict: The source spec dict to minimize.
    :type spec_dict: dict
    :param operations: A list of operation IDs to retain.
    :type operations: list of str
    :param resources: A list of resource names to retain.
    :type resources: list of str
    :return: Minimized swagger spec dict
    :rtype: dict
    """
    operations = operations or []
    resources = resources or []

    # keep the ugly overhead for now but only add paths we need
    minimized = {key: value for key, value in spec_dict.items() if key != 'paths'}
    minimized['paths'] = {}

    for path_name, path in spec_dict['paths'].items():
        for method, data in path.items():
            if (
                data['operationId'] in operations
                or any(tag in resources for tag in data['tags'])
            ):
                if path_name not in minimized['paths']:
                    minimized['paths'][path_name] = {}
                minimized['paths'][path_name][method] = data

    return minimized


class EsiClientProvider:
    """Class for providing a single ESI client instance for the whole app

    Args:
        datasource: Name of the ESI datasource to access.
        spec_file: Absolute path to a swagger spec file to load.
        version: Base ESI API version. Accepted values are 'legacy', 'latest',
        ua_appname: Name of the App for generating a User-Agent,
        ua_version: Version of the App for generating a User-Agent,
        ua_url: (optional) URL To the Source Code or Documentation for generating a User-Agent,
        kwargs: Explicit resource versions to build, in the form Character='v4'. \
            Same values accepted as version.

    If a spec_file is specified, specific versioning is not available.
    Meaning the version and resource version kwargs are ignored in favour of the
    versions available in the spec_file.
    """

    _client = None

    def __init__(
            self, datasource=None, spec_file=None, version=None,
            app_info_text=None,  # Deprecate in favour of the following variables
            ua_appname: str = None, ua_version: str = None, ua_url: str = None, **kwargs) -> None:
        self._datasource = datasource
        self._spec_file = spec_file
        self._version = version
        self._app_text = app_info_text  # Deprecate in favour of the following variables
        self._ua_appname = ua_appname
        self._ua_version = ua_version
        self._ua_url = ua_url
        self._kwargs = kwargs

        if app_info_text is not None:
            warnings.warn(
                "The 'app_info_text' parameter is deprecated and will be removed in a future release. "
                "Use 'ua_appname', 'ua_version', and `ua_url` to dynamically build a User-Agent instead",
                DeprecationWarning,
                stacklevel=2
            )

    @property
    def client(self) -> SwaggerClient:
        if self._client is None:
            self._client = esi_client_factory(
                datasource=self._datasource,
                spec_file=self._spec_file,
                version=self._version,
                app_info_text=self._app_text,  # Deprecate in favour of the following variables
                ua_appname=self._ua_appname,
                ua_version=self._ua_version,
                ua_url=self._ua_url,
                **self._kwargs,
            )
        return self._client

    def __str__(self) -> str:
        return 'EsiClientProvider'



"""Django app for accessing the EVE Swagger Interface (ESI)."""

__version__ = "8.2.5"
__title__ = 'Django-ESI'
__url__ = 'https://gitlab.com/allianceauth/django-esi'
__build_date__ = "2025-12-16"
__esi_compatibility_date__ = "2025-12-16"



import datetime
import os
from unittest.mock import patch, Mock
import json

import bravado
from bravado_core.spec import Spec
from bravado.requests_client import RequestsClient
from bravado.exception import HTTPBadGateway
import requests_mock

import django
from django.contrib.auth.models import User
from django.core.cache import cache

from . import _generate_token, _store_as_Token, NoSocketsTestCase
from .factories import create_http_error
from ..clients import (
    EsiClientProvider,
    esi_client_factory,
    TokenAuthenticator,
    build_cache_name,
    build_spec,
    build_spec_url,
    cache_spec,
    get_spec,
    read_spec,
    minimize_spec,
    SwaggerClient,
    CachingHttpFuture,
    RequestsClientPlus,
)
from ..errors import TokenExpiredError

SWAGGER_SPEC_PATH_MINIMAL = os.path.join(
    os.path.dirname(os.path.abspath(__file__)), "test_swagger.json"
)
SWAGGER_SPEC_PATH_FULL = os.path.join(
    os.path.dirname(os.path.abspath(__file__)), "test_swagger_full.json"
)

MODULE_PATH = "esi.clients"


def _load_json_file(path):
    with open(path, encoding="utf-8") as f:
        return json.load(f)


class MockResultFuture:
    def __init__(self):
        dt = datetime.datetime.now(datetime.timezone.utc) + datetime.timedelta(seconds=60)
        self.headers = {"Expires": dt.strftime("%a, %d %b %Y %H:%M:%S %Z")}
        self.status_code = 200
        self.text = "dummy"


class MockResultPast:
    def __init__(self):
        dt = datetime.datetime.now(datetime.timezone.utc) - datetime.timedelta(seconds=60)
        self.headers = {"Expires": dt.strftime("%a, %d %b %Y %H:%M:%S %Z")}


@patch.object(django.core.cache.cache, "set")
@patch.object(django.core.cache.cache, "get")
@patch.object(bravado.http_future.HttpFuture, "result")
class TestClientCache(NoSocketsTestCase):
    @classmethod
    def setUpTestData(cls):
        spec = _load_json_file(SWAGGER_SPEC_PATH_MINIMAL)
        with requests_mock.Mocker() as requests_mocker:
            requests_mocker.register_uri(
                "GET", url="https://esi.evetech.net/_latest/swagger.json", json=spec
            )
            cls.c = esi_client_factory(spec_file=SWAGGER_SPEC_PATH_MINIMAL)

    def setUp(self) -> None:
        cache.clear()

    def test_cache_expire(self, mock_future_result, mock_cache_get, mock_cache_set):
        mock_future_result.return_value = ({"players": 500}, MockResultFuture())
        mock_cache_get.return_value = False

        # hit api
        r = self.c.Status.get_status().result()
        self.assertEqual(r["players"], 500)

        mock_cache_get.return_value = ({"players": 50}, MockResultFuture())
        # hit cache and pass
        r = self.c.Status.get_status().result()
        self.assertEqual(r["players"], 50)

        mock_cache_get.return_value = ({"players": 50}, MockResultPast())
        # hit cache fail, re-hit api
        r = self.c.Status.get_status().result()
        self.assertEqual(r["players"], 500)

    def test_should_use_cache(self, mock_future_result, mock_cache_get, mock_cache_set):
        # given
        mock_future_result.return_value = ({"players": 500}, MockResultFuture())
        mock_cache_get.return_value = False

        # when
        r = self.c.Status.get_status().result()
        # then
        self.assertEqual(r["players"], 500)
        self.assertTrue(mock_cache_get.called)
        self.assertTrue(mock_cache_set.called)

    def test_should_not_use_cache(
        self, mock_future_result, mock_cache_get, mock_cache_set
    ):
        # given
        mock_future_result.return_value = ({"players": 500}, MockResultFuture())
        mock_cache_get.return_value = False

        # when
        r = self.c.Status.get_status().result(ignore_cache=True)
        # then
        self.assertEqual(r["players"], 500)
        self.assertFalse(mock_cache_get.called)
        self.assertFalse(mock_cache_set.called)

    def test_can_handle_exception_from_cache_set(
        self, mock_future_result, mock_cache_get, mock_cache_set
    ):
        mock_future_result.return_value = ({"players": 500}, MockResultFuture())
        mock_cache_get.return_value = False
        mock_cache_set.side_effect = RuntimeError("TEST: Could not write to cache")

        # hit api
        r = self.c.Status.get_status().result()
        self.assertEqual(r["players"], 500)

    def test_can_handle_exception_from_cache_get(
        self, mock_future_result, mock_cache_get, mock_cache_set
    ):
        mock_future_result.return_value = ({"players": 500}, MockResultFuture())
        mock_cache_get.side_effect = RuntimeError("TEST: Could not read from cache")

        # hit api
        r = self.c.Status.get_status().result()
        self.assertEqual(r["players"], 500)


@patch(MODULE_PATH + ".app_settings.ESI_SPEC_CACHE_DURATION", 3)
@patch(MODULE_PATH + ".app_settings.ESI_API_URL", "https://www.example.com/esi/")
@patch(MODULE_PATH + ".app_settings.ESI_API_DATASOURCE", "dummy")
@patch("esi.models.app_settings.ESI_TOKEN_VALID_DURATION", 120)
class TestTokenAuthenticator(NoSocketsTestCase):
    def setUp(self):
        self.user = User.objects.create_user(
            "Bruce Wayne", "abc@example.com", "password"
        )
        self.token = _store_as_Token(
            _generate_token(
                character_id=101,
                character_name=self.user.username,
                scopes=["abc"],
                access_token="my_access_token",
            ),
            self.user,
        )

    def test_apply_defaults(self):
        request = Mock()
        request.headers = dict()
        request.params = dict()

        x = TokenAuthenticator()
        request2 = x.apply(request)
        self.assertEqual(request2.headers["Authorization"], None)
        self.assertEqual(request2.params["datasource"], "dummy")

    def test_apply_token(self):
        request = Mock()
        request.headers = dict()
        request.params = dict()

        x = TokenAuthenticator(token=self.token)
        request2 = x.apply(request)
        self.assertEqual(request2.headers["Authorization"], "Bearer my_access_token")
        self.assertEqual(request2.params["datasource"], "dummy")

    def test_apply_token_datasource(self):
        request = Mock()
        request.headers = dict()
        request.params = dict()

        x = TokenAuthenticator(token=self.token, datasource="dummy2")
        request2 = x.apply(request)
        self.assertEqual(request2.headers["Authorization"], "Bearer my_access_token")
        self.assertEqual(request2.params["datasource"], "dummy2")

    @patch("esi.models.Token.refresh", spec=True)
    def test_apply_token_expired_success(self, mock_Token_refresh):
        request = Mock()
        request.headers = dict()
        request.params = dict()

        self.token.created -= datetime.timedelta(121)

        x = TokenAuthenticator(token=self.token)
        request2 = x.apply(request)
        self.assertEqual(request2.headers["Authorization"], "Bearer my_access_token")
        self.assertEqual(request2.params["datasource"], "dummy")
        self.assertEqual(mock_Token_refresh.call_count, 1)

    @patch("esi.models.Token.refresh", spec=True)
    def test_apply_token_expired_failed(self, mock_Token_refresh):
        request = Mock()
        request.headers = dict()
        request.params = dict()

        self.token.created -= datetime.timedelta(121)
        self.token.refresh_token = None

        x = TokenAuthenticator(token=self.token)
        with self.assertRaises(TokenExpiredError):
            x.apply(request)

        self.assertEqual(mock_Token_refresh.call_count, 0)


class TestModuleFunctions(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.spec = _load_json_file(SWAGGER_SPEC_PATH_MINIMAL)

    def test_build_cache_name(self):
        self.assertEqual(build_cache_name("abc"), "esi_swaggerspec_abc")

    @patch(MODULE_PATH + ".app_settings.ESI_SPEC_CACHE_DURATION", 3)
    def test_cache_spec(self):
        spec = {"dummy_spec": True}
        cache_spec("abc", spec)
        self.assertDictEqual(cache.get("esi_swaggerspec_abc"), spec)

    @patch(MODULE_PATH + ".app_settings.ESI_API_URL", "https://www.example.com/esi/")
    def test_build_spec_url(self):
        self.assertEqual(
            build_spec_url("v2"), "https://www.example.com/esi/v2/swagger.json"
        )

    @requests_mock.Mocker()
    @patch(MODULE_PATH + ".app_settings.ESI_SPEC_CACHE_DURATION", 1)
    def test_get_spec_defaults(self, requests_mocker):
        # given
        requests_mocker.register_uri(
            "GET", url="https://esi.evetech.net/_latest/swagger.json", json=self.spec
        )
        requests_mocker.register_uri(
            "GET", url="https://esi.evetech.net/latest/swagger.json", json=self.spec
        )
        # when
        spec = get_spec("latest")
        # then
        self.assertIsInstance(spec, Spec)

    @patch(MODULE_PATH + ".app_settings.ESI_SPEC_CACHE_DURATION", 1)
    def test_get_spec_with_http_client(self):
        mock_http_client = Mock(spec=RequestsClient)
        mock_http_client.request.return_value.result.return_value.json.return_value = (
            self.spec
        )
        spec = get_spec("latest", http_client=mock_http_client)
        self.assertIsInstance(spec, Spec)

    @patch(MODULE_PATH + ".app_settings.ESI_SPEC_CACHE_DURATION", 1)
    def test_get_spec_with_config(self):
        mock_http_client = Mock(spec=RequestsClient)
        mock_http_client.request.return_value.result.return_value.json.return_value = (
            self.spec
        )
        spec = get_spec(
            "latest", http_client=mock_http_client, config={"dummy_config": True}
        )
        self.assertIsInstance(spec, Spec)
        self.assertIn("dummy_config", spec.config)

    @patch(MODULE_PATH + ".app_settings.ESI_SPEC_CACHE_DURATION", 1)
    def test_build_spec_defaults(self):
        mock_http_client = Mock(spec=RequestsClient)
        mock_http_client.request.return_value.result.return_value.json.return_value = (
            self.spec
        )
        spec = build_spec("v1", http_client=mock_http_client)
        self.assertIsInstance(spec, Spec)

    @patch(MODULE_PATH + ".app_settings.ESI_SPEC_CACHE_DURATION", 1)
    def test_build_spec_explicit_resource_found(self):
        mock_http_client = Mock(spec=RequestsClient)
        mock_http_client.request.return_value.result.return_value.json.return_value = (
            self.spec
        )
        spec = build_spec("v1", http_client=mock_http_client, Status="v1")
        self.assertIsInstance(spec, Spec)

    @patch(MODULE_PATH + ".app_settings.ESI_SPEC_CACHE_DURATION", 1)
    def test_build_spec_explicit_resource_not_found(self):
        mock_http_client = Mock(spec=RequestsClient)
        mock_http_client.request.return_value.result.return_value.json.return_value = (
            self.spec
        )
        with self.assertRaises(AttributeError):
            build_spec("v1", http_client=mock_http_client, Character="v4")

    @requests_mock.Mocker()
    def test_read_spec(self, requests_mocker):
        # given
        requests_mocker.register_uri(
            "GET", url="https://esi.evetech.net/_latest/swagger.json", json=self.spec
        )
        # when
        client = read_spec(SWAGGER_SPEC_PATH_MINIMAL)
        # then
        self.assertIsInstance(client, SwaggerClient)

    def test_minimize_spec_defaults(self):
        spec_dict = minimize_spec(self.spec)
        self.assertIsInstance(spec_dict, dict)
        # todo: add better verification of functionality

    def test_minimize_spec_resources(self):
        spec_dict = minimize_spec(self.spec, resources=["Status"])
        self.assertIsInstance(spec_dict, dict)
        # todo: add better verification of functionality


@patch(MODULE_PATH + ".app_settings.ESI_SPEC_CACHE_DURATION", 1)
@requests_mock.Mocker()
class TestEsiClientFactory(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.spec = _load_json_file(SWAGGER_SPEC_PATH_MINIMAL)

    def setUp(self):
        self.user = User.objects.create_user(
            "Bruce Wayne", "abc@example.com", "password"
        )
        self.token = _store_as_Token(
            _generate_token(
                character_id=101,
                character_name=self.user.username,
                scopes=["abc"],
                access_token="my_access_token",
            ),
            self.user,
        )

    def test_minimal_client(self, requests_mocker):
        requests_mocker.register_uri(
            "GET", url="https://esi.evetech.net/_latest/swagger.json", json=self.spec
        )
        requests_mocker.register_uri(
            "GET", url="https://esi.evetech.net/latest/swagger.json", json=self.spec
        )
        # when
        client = esi_client_factory()
        # then
        self.assertIsInstance(client, SwaggerClient)

    def test_client_with_token(self, requests_mocker):
        # given
        requests_mocker.register_uri(
            "GET", url="https://esi.evetech.net/_latest/swagger.json", json=self.spec
        )
        requests_mocker.register_uri(
            "GET", url="https://esi.evetech.net/latest/swagger.json", json=self.spec
        )
        # when
        client = esi_client_factory(token=self.token)
        # then
        self.assertIsInstance(client, SwaggerClient)

    def test_client_with_version(self, requests_mocker):
        # given
        requests_mocker.register_uri(
            "GET", url="https://esi.evetech.net/latest/swagger.json", json=self.spec
        )
        requests_mocker.register_uri(
            "GET", url="https://esi.evetech.net/_latest/swagger.json", json=self.spec
        )
        requests_mocker.register_uri(
            "GET", url="https://esi.evetech.net/v1/swagger.json", json=self.spec
        )
        # when
        client = esi_client_factory(version="v1")
        # then
        self.assertIsInstance(client, SwaggerClient)

    def test_client_with_spec_file(self, requests_mocker):
        # given
        requests_mocker.register_uri(
            "GET", url="https://esi.evetech.net/latest/swagger.json", json=self.spec
        )
        requests_mocker.register_uri(
            "GET", url="https://esi.evetech.net/_latest/swagger.json", json=self.spec
        )
        # when
        client = esi_client_factory(spec_file=SWAGGER_SPEC_PATH_MINIMAL)
        # then
        self.assertIsInstance(client, SwaggerClient)

    def test_client_with_explicit_resource(self, requests_mocker):
        # given
        requests_mocker.register_uri(
            "GET", url="https://esi.evetech.net/latest/swagger.json", json=self.spec
        )
        requests_mocker.register_uri(
            "GET", url="https://esi.evetech.net/_latest/swagger.json", json=self.spec
        )
        requests_mocker.register_uri(
            "GET", url="https://esi.evetech.net/v1/swagger.json", json=self.spec
        )
        # when
        client = esi_client_factory(Status="v1")
        # then
        self.assertIsInstance(client, SwaggerClient)

    def test__time_to_expiry_failure(self, requests_mocker):
        seconds = CachingHttpFuture._time_to_expiry("fail")
        self.assertEqual(seconds, 0)


@patch(MODULE_PATH + ".HttpFuture.result")
class TestClientResult(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        spec = _load_json_file(SWAGGER_SPEC_PATH_MINIMAL)
        with requests_mock.Mocker() as requests_mocker:
            requests_mocker.register_uri(
                "GET", url="https://esi.evetech.net/_latest/swagger.json", json=spec
            )
            cls.esi_client = esi_client_factory(spec_file=SWAGGER_SPEC_PATH_MINIMAL)

    @patch(MODULE_PATH + ".app_settings.ESI_REQUESTS_CONNECT_TIMEOUT", 10)
    @patch(MODULE_PATH + ".app_settings.ESI_REQUESTS_READ_TIMEOUT", 60)
    def test_use_default_timeout(self, mock_future_result):
        # given
        mock_future_result.return_value = (None, Mock(**{"headers": {}}))
        # when
        self.esi_client.Status.get_status().result()
        # then
        self.assertTrue(mock_future_result.called)
        _, kwargs = mock_future_result.call_args
        self.assertEqual(kwargs["timeout"], (10, 60))

    @patch(MODULE_PATH + ".app_settings.ESI_REQUESTS_CONNECT_TIMEOUT", 10)
    @patch(MODULE_PATH + ".app_settings.ESI_REQUESTS_READ_TIMEOUT", 60)
    def test_use_custom_timeout(self, mock_future_result):
        # given
        mock_future_result.return_value = (None, Mock(**{"headers": {}}))
        # when
        self.esi_client.Status.get_status().result(timeout=42)
        # then
        self.assertTrue(mock_future_result.called)
        _, kwargs = mock_future_result.call_args
        self.assertEqual(kwargs["timeout"], 42)

    def test_support_language_parameter(self, mock_future_result):
        # given
        mock_future_result.return_value = (None, Mock(**{"headers": {}}))
        my_language = "de"
        operation = self.esi_client.Status.get_status()
        # when
        operation.result(language=my_language)
        # then
        self.assertTrue(mock_future_result.called)
        self.assertEqual(operation.future.request.params["language"], my_language)
        _, kwargs = mock_future_result.call_args
        self.assertNotIn("language", kwargs)

    @patch(MODULE_PATH + ".app_settings.ESI_SERVER_ERROR_BACKOFF_FACTOR", 0.5)
    @patch(MODULE_PATH + ".app_settings.ESI_SERVER_ERROR_MAX_RETRIES", 3)
    @patch(MODULE_PATH + ".sleep", wraps=lambda x: None)
    def test_retries_1(self, mock_sleep, mock_future_result):
        mock_future_result.side_effect = create_http_error(502)
        try:
            self.esi_client.Status.get_status().result()
        except HTTPBadGateway as e:
            # requests error thrown
            self.assertIsInstance(e, HTTPBadGateway)
            # we tried # times before raising
            self.assertEqual(mock_future_result.call_count, 4)
            call_list = mock_sleep.call_args_list
            result = [args[0] for args, _ in [x for x in call_list]]
            expected = [0.5, 1.0, 2.0]
            self.assertListEqual(expected, result)

    @patch(MODULE_PATH + ".app_settings.ESI_SERVER_ERROR_BACKOFF_FACTOR", 0.5)
    @patch(MODULE_PATH + ".app_settings.ESI_SERVER_ERROR_MAX_RETRIES", 1)
    @patch(MODULE_PATH + ".sleep", lambda x: None)
    def test_retries_2(self, mock_future_result):
        mock_future_result.side_effect = create_http_error(502)
        try:
            self.esi_client.Status.get_status().result()
        except HTTPBadGateway as e:
            # requests error thrown
            self.assertIsInstance(e, HTTPBadGateway)
            # we tried # times before raising
            self.assertEqual(mock_future_result.call_count, 2)

    @patch(MODULE_PATH + ".app_settings.ESI_SERVER_ERROR_BACKOFF_FACTOR", 0.5)
    @patch(MODULE_PATH + ".app_settings.ESI_SERVER_ERROR_MAX_RETRIES", 0)
    @patch(MODULE_PATH + ".sleep", lambda x: None)
    def test_retries_3(self, mock_future_result):
        mock_future_result.side_effect = create_http_error(502)
        try:
            self.esi_client.Status.get_status().result()
        except HTTPBadGateway as e:
            # requests error thrown
            self.assertIsInstance(e, HTTPBadGateway)
            # we tried # times before raising
            self.assertEqual(mock_future_result.call_count, 1)

    @patch(MODULE_PATH + ".app_settings.ESI_SERVER_ERROR_BACKOFF_FACTOR", 0.5)
    @patch(MODULE_PATH + ".app_settings.ESI_SERVER_ERROR_MAX_RETRIES", 4)
    @patch(MODULE_PATH + ".sleep", lambda x: None)
    def test_retry_with_custom_retries(self, mock_future_result):
        mock_future_result.side_effect = create_http_error(502)
        try:
            self.esi_client.Status.get_status().result(retries=1)
        except HTTPBadGateway as e:
            # requests error thrown
            self.assertIsInstance(e, HTTPBadGateway)
            # we tried # times before raising
            self.assertEqual(mock_future_result.call_count, 2)


@patch(MODULE_PATH + ".HttpFuture.result")
class TestClientResultAllPages(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        spec = _load_json_file(SWAGGER_SPEC_PATH_FULL)
        with requests_mock.Mocker() as requests_mocker:
            requests_mocker.register_uri(
                "GET", url="https://esi.evetech.net/_latest/swagger.json", json=spec
            )
            cls.esi_client = esi_client_factory(spec_file=SWAGGER_SPEC_PATH_FULL)

    def test_pages(self, mock_future_result):
        class MockResultHeaders:
            def __init__(self):
                self.headers = {"X-Pages": 10}
                self.status_code = 200
                self.text = "dummy"

        # given
        mock_future_result.return_value = ({"contract_test": 1}, MockResultHeaders())
        self.esi_client.Contracts.get_contracts_public_region_id(region_id=1).results()
        # then
        self.assertEqual(mock_future_result.call_count, 10)  # we got 10 pages of data

    def test_pages_response(self, mock_future_result):
        class MockResultHeaders:
            def __init__(self):
                self.headers = {"X-Pages": 10}
                self.status_code = 200
                self.text = "dummy"

        # given
        mock_future_result.return_value = ({"contract_test": 1}, MockResultHeaders())
        o = self.esi_client.Contracts.get_contracts_public_region_id(region_id=1)
        o.request_config.also_return_response = True
        # when
        result, response = o.results()
        # then
        self.assertEqual(mock_future_result.call_count, 10)  # we got 10 pages of data
        self.assertEqual(len(result), 10)  # we got 10 lots of data
        self.assertEqual(response.headers, {"X-Pages": 10})  # we got header of data

    def test_pages_on_non_paged_endpoint(self, mock_future_result):
        class MockResultHeaders:
            def __init__(self):
                self.headers = {"header_test": "ok"}
                self.status_code = 200
                self.text = "dummy"

        # given
        mock_future_result.return_value = ({"status_test": 1}, MockResultHeaders())
        # when
        self.esi_client.Status.get_status().results()
        # then
        self.assertEqual(mock_future_result.call_count, 1)  # we got no pages of data


@patch(MODULE_PATH + ".app_settings.ESI_LANGUAGES", ["lang1", "lang2", "lang3"])
@patch(MODULE_PATH + ".CachingHttpFuture.results", spec=True)
@requests_mock.Mocker()
class TestClientResultsLocalized(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        cls.spec = _load_json_file(SWAGGER_SPEC_PATH_MINIMAL)

    @staticmethod
    def my_results(**kwargs):
        if "language" in kwargs:
            return "response_" + kwargs["language"]
        return ""

    def test_default(self, mock_future_results, requests_mocker):
        # given
        mock_future_results.side_effect = self.my_results
        requests_mocker.register_uri(
            "GET", url="https://esi.evetech.net/_latest/swagger.json", json=self.spec
        )
        client = esi_client_factory(spec_file=SWAGGER_SPEC_PATH_MINIMAL)
        # when
        result = client.Status.get_status().results_localized()
        # then
        expected = {
            "lang1": "response_lang1",
            "lang2": "response_lang2",
            "lang3": "response_lang3",
        }
        self.assertDictEqual(result, expected)

    def test_custom_languages(self, mock_future_results, requests_mocker):
        # given
        mock_future_results.side_effect = self.my_results
        requests_mocker.register_uri(
            "GET", url="https://esi.evetech.net/_latest/swagger.json", json=self.spec
        )
        client = esi_client_factory(spec_file=SWAGGER_SPEC_PATH_MINIMAL)
        # when
        result = client.Status.get_status().results_localized(
            languages=["lang2", "lang3"]
        )
        # then
        expected = {
            "lang2": "response_lang2",
            "lang3": "response_lang3",
        }
        self.assertDictEqual(result, expected)

    def test_raise_on_invalid_language(self, mock_future_results, requests_mocker):
        # given
        mock_future_results.side_effect = self.my_results
        requests_mocker.register_uri(
            "GET", url="https://esi.evetech.net/_latest/swagger.json", json=self.spec
        )
        client = esi_client_factory(spec_file=SWAGGER_SPEC_PATH_MINIMAL)
        # when/then
        with self.assertRaises(ValueError):
            client.Status.get_status().results_localized(languages=["lang2", "xxx"])


@requests_mock.Mocker()
class TestEsiClientProvider(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        cls.spec = _load_json_file(SWAGGER_SPEC_PATH_MINIMAL)

    def test_client_loads_on_demand(self, requests_mocker):
        # given
        requests_mocker.register_uri(
            "GET", url="https://esi.evetech.net/_latest/swagger.json", json=self.spec
        )
        requests_mocker.register_uri(
            "GET", url="https://esi.evetech.net/latest/swagger.json", json=self.spec
        )
        my_provider = EsiClientProvider()
        # when
        esi_client = my_provider.client
        # then
        self.assertIsInstance(esi_client, SwaggerClient)
        self.assertIsInstance(my_provider, EsiClientProvider)
        self.assertIsInstance(my_provider.client, SwaggerClient)
        self.assertEqual(str(my_provider), "EsiClientProvider")

    # FIXME: This test currently fails sometimes when run with tox.
    # def test_with_version(self, requests_mocker):
    #     # given
    #     requests_mocker.register_uri(
    #         "GET", url="https://esi.evetech.net/_latest/swagger.json", json=self.spec
    #     )
    #     requests_mocker.register_uri(
    #         "GET", url="https://esi.evetech.net/latest/swagger.json", json=self.spec
    #     )
    #     requests_mocker.register_uri(
    #         "GET", url="https://esi.evetech.net/v1/swagger.json", json=self.spec
    #     )
    #     my_provider = EsiClientProvider(version="v1")
    #     # when
    #     esi_client = my_provider.client
    #     # then
    #     self.assertIsInstance(esi_client, SwaggerClient)
    #     urls_callled = [x.url for x in requests_mocker.request_history]
    #     self.assertIn("https://esi.evetech.net/v1/swagger.json", urls_callled)

    def test_with_spec_file(self, requests_mocker):
        # given
        requests_mocker.register_uri(
            "GET", url="https://esi.evetech.net/_latest/swagger.json", json=self.spec
        )
        requests_mocker.register_uri(
            "GET", url="https://esi.evetech.net/latest/swagger.json", json=self.spec
        )
        my_provider = EsiClientProvider(spec_file=SWAGGER_SPEC_PATH_MINIMAL)
        # when
        esi_client = my_provider.client
        # then
        self.assertIsInstance(esi_client, SwaggerClient)


@requests_mock.Mocker()
class TestClientResult2(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        spec = _load_json_file(SWAGGER_SPEC_PATH_MINIMAL)
        with patch(MODULE_PATH + ".app_settings.ESI_USER_CONTACT_EMAIL", "email@example.com"), patch(
            MODULE_PATH + ".__title__", "Django-ESI"
        ), patch(MODULE_PATH + ".__version__", "1.0.0"):
            with requests_mock.Mocker() as requests_mocker:
                requests_mocker.register_uri(
                    "GET", url="https://esi.evetech.net/_latest/swagger.json", json=spec
                )
                cls.esi_client = esi_client_factory(spec_file=SWAGGER_SPEC_PATH_MINIMAL)

    def test_normal_call(self, requests_mocker):
        # given
        requests_mocker.register_uri("GET", url="https://esi.evetech.net/v1/status/")
        # when
        self.esi_client.Status.get_status().result()
        # then
        self.assertTrue(requests_mocker.called)
        request = requests_mocker.last_request

        expected_title = 'DjangoEsi'

        self.assertEqual(request._request.headers["User-Agent"], f"{expected_title}/1.0.0 (email@example.com; +https://gitlab.com/allianceauth/django-esi)")

    def test_existing_headers(self, requests_mocker):
        # given
        requests_mocker.register_uri("GET", url="https://esi.evetech.net/v1/status/")
        # when
        self.esi_client.Status.get_status().result()
        # then
        self.assertTrue(requests_mocker.called)
        request = requests_mocker.last_request

        expected_title = 'DjangoEsi'

        self.assertEqual(request._request.headers["User-Agent"], f"{expected_title}/1.0.0 (email@example.com; +https://gitlab.com/allianceauth/django-esi)")


class TestRequestsClientPlus(NoSocketsTestCase):
    def test_single_header(self):
        """When no headers are set, just add the new header"""
        obj = RequestsClientPlus()
        obj.user_agent = "abc"
        result = obj.request(
            {"method": "GET", "url": "https://esi.evetech.net/v1/status/"}
        )
        self.assertEqual(result.future.request.headers["User-Agent"], "abc")

    def test_existing_header(self):
        """When a header exists, leave it intact"""
        obj = RequestsClientPlus()
        obj.user_agent = "abc"
        result = obj.request(
            {
                "method": "GET",
                "url": "https://esi.evetech.net/v1/status/",
                "headers": {"From": "email@example.com"},
            }
        )
        self.assertEqual(result.future.request.headers["User-Agent"], "abc")
        self.assertEqual(result.future.request.headers["From"], "email@example.com")

    def test_no_user_agent(self):
        """When no user agent is defined, leave the existing header intact"""
        obj = RequestsClientPlus()
        result = obj.request(
            {
                "method": "GET",
                "url": "https://esi.evetech.net/v1/status/",
                "headers": {"From": "email@example.com"},
            }
        )
        self.assertEqual(result.future.request.headers["From"], "email@example.com")
        self.assertNotIn("User-Agent", result.future.request.headers)


@patch(MODULE_PATH + ".__title__", "Django-ESI")
@patch(MODULE_PATH + ".__version__", "1.0.0")
@requests_mock.Mocker()
class TestEsiClientFactoryAppText(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        cls.spec = _load_json_file(SWAGGER_SPEC_PATH_MINIMAL)
        cls.status_response = {
            "players": 12345,
            "server_version": "1132976",
            "start_time": "2017-01-02T12:34:56Z",
        }

    @patch(MODULE_PATH + ".app_settings.ESI_USER_CONTACT_EMAIL", None)
    def test_defaults(self, requests_mocker) -> None:
        # This test is not expected to hit given that ESI_USER_CONTACT_EMAIL must be set
        # But here it is for completeness
        requests_mocker.register_uri("GET", url="https://esi.evetech.net/_latest/swagger.json", json=self.spec)
        requests_mocker.register_uri("GET", url="https://esi.evetech.net/latest/swagger.json", json=self.spec)
        requests_mocker.register_uri("GET", url="https://esi.evetech.net/v1/status/", json=self.status_response)
        client = esi_client_factory()
        # when
        operation = client.Status.get_status()
        # then
        expected_app_name = "MyApp"
        expected_title = 'DjangoEsi'
        self.assertEqual(operation.future.request.headers["User-Agent"], f"{expected_title}/1.0.0 (None; +https://gitlab.com/allianceauth/django-esi)")

    @patch(MODULE_PATH + ".app_settings.ESI_USER_CONTACT_EMAIL", "email@example.com")
    def test_defaults_email(self, requests_mocker) -> None:
        # given
        requests_mocker.register_uri("GET", url="https://esi.evetech.net/_latest/swagger.json", json=self.spec)
        requests_mocker.register_uri("GET", url="https://esi.evetech.net/latest/swagger.json", json=self.spec)
        requests_mocker.register_uri("GET", url="https://esi.evetech.net/v1/status/", json=self.status_response)
        client = esi_client_factory()
        # when
        operation = client.Status.get_status()
        # then
        expected_app_name = "MyApp"
        expected_title = 'DjangoEsi'
        self.assertEqual(operation.future.request.headers["User-Agent"], f"{expected_title}/1.0.0 (email@example.com; +https://gitlab.com/allianceauth/django-esi)")

    @patch(MODULE_PATH + ".app_settings.ESI_USER_CONTACT_EMAIL", None)
    def test_app_text(self, requests_mocker) -> None:
        # Deprecated
        # This test is not expected to hit given that ESI_USER_CONTACT_EMAIL must be set
        # given
        requests_mocker.register_uri("GET", url="https://esi.evetech.net/_latest/swagger.json", json=self.spec)
        requests_mocker.register_uri("GET", url="https://esi.evetech.net/latest/swagger.json", json=self.spec)
        requests_mocker.register_uri("GET", url="https://esi.evetech.net/v1/status/", json=self.status_response)
        client = esi_client_factory(app_info_text="my-app v1.0.0")
        # when
        operation = client.Status.get_status()
        # then
        expected_app_name = "MyApp"
        expected_title = 'DjangoEsi'
        self.assertEqual(operation.future.request.headers["User-Agent"], f"my-app v1.0.0 (None) {expected_title}/1.0.0 (+https://gitlab.com/allianceauth/django-esi)",)

    @patch(MODULE_PATH + ".app_settings.ESI_USER_CONTACT_EMAIL", "email@example.com")
    def test_app_text_with_email(self, requests_mocker):
        # Deprecated
        # given
        requests_mocker.register_uri("GET", url="https://esi.evetech.net/_latest/swagger.json", json=self.spec)
        requests_mocker.register_uri("GET", url="https://esi.evetech.net/latest/swagger.json", json=self.spec)
        requests_mocker.register_uri("GET", url="https://esi.evetech.net/v1/status/", json=self.status_response)
        client = esi_client_factory(app_info_text="my-app v1.0.0")
        # when
        operation = client.Status.get_status()
        # then
        expected_app_name = "MyApp"
        expected_title = 'DjangoEsi'
        self.assertEqual(operation.future.request.headers["User-Agent"], f"my-app v1.0.0 (email@example.com) {expected_title}/1.0.0 (+https://gitlab.com/allianceauth/django-esi)",)

    @patch(MODULE_PATH + ".app_settings.ESI_USER_CONTACT_EMAIL", "email@example.com")
    def test_ua_generator(self, requests_mocker):
        requests_mocker.register_uri("GET", url="https://esi.evetech.net/_latest/swagger.json", json=self.spec)
        requests_mocker.register_uri("GET", url="https://esi.evetech.net/latest/swagger.json", json=self.spec)
        requests_mocker.register_uri("GET", url="https://esi.evetech.net/v1/status/", json=self.status_response)
        client = esi_client_factory(ua_appname="MyApp", ua_version="1.0.0")
        # when
        operation = client.Status.get_status()
        # then
        expected_app_name = "MyApp"
        expected_title = 'DjangoEsi'
        self.assertEqual(operation.future.request.headers["User-Agent"], f"{expected_app_name}/1.0.0 (email@example.com) {expected_title}/1.0.0 (+https://gitlab.com/allianceauth/django-esi)")

    @patch(MODULE_PATH + ".app_settings.ESI_USER_CONTACT_EMAIL", "email@example.com")
    def test_ua_generator_for_appname_with_spaces(self, requests_mocker):
        requests_mocker.register_uri("GET", url="https://esi.evetech.net/_latest/swagger.json", json=self.spec)
        requests_mocker.register_uri("GET", url="https://esi.evetech.net/latest/swagger.json", json=self.spec)
        requests_mocker.register_uri("GET", url="https://esi.evetech.net/v1/status/", json=self.status_response)
        client = esi_client_factory(ua_appname="My App", ua_version="1.0.0")

        # when
        operation = client.Status.get_status()

        # then
        expected_app_name = "MyApp"
        expected_title = 'DjangoEsi'

        self.assertEqual(operation.future.request.headers["User-Agent"], f"{expected_app_name}/1.0.0 (email@example.com) {expected_title}/1.0.0 (+https://gitlab.com/allianceauth/django-esi)")

    @patch(MODULE_PATH + ".app_settings.ESI_USER_CONTACT_EMAIL", "email@example.com")
    def test_ua_generator_for_appname_with_hyphens(self, requests_mocker):
        requests_mocker.register_uri("GET", url="https://esi.evetech.net/_latest/swagger.json", json=self.spec)
        requests_mocker.register_uri("GET", url="https://esi.evetech.net/latest/swagger.json", json=self.spec)
        requests_mocker.register_uri("GET", url="https://esi.evetech.net/v1/status/", json=self.status_response)
        client = esi_client_factory(ua_appname="My-App", ua_version="1.0.0")

        # when
        operation = client.Status.get_status()

        # then
        expected_app_name = "MyApp"
        expected_title = 'DjangoEsi'

        self.assertEqual(operation.future.request.headers["User-Agent"], f"{expected_app_name}/1.0.0 (email@example.com) {expected_title}/1.0.0 (+https://gitlab.com/allianceauth/django-esi)")

    @patch(MODULE_PATH + ".app_settings.ESI_USER_CONTACT_EMAIL", "email@example.com")
    def test_ua_generator_with_url(self, requests_mocker):
        requests_mocker.register_uri("GET", url="https://esi.evetech.net/_latest/swagger.json", json=self.spec)
        requests_mocker.register_uri("GET", url="https://esi.evetech.net/latest/swagger.json", json=self.spec)
        requests_mocker.register_uri("GET", url="https://esi.evetech.net/v1/status/", json=self.status_response)
        client = esi_client_factory(ua_appname="MyApp", ua_version="1.0.0", ua_url="https://example.com")
        # when
        operation = client.Status.get_status()
        # then
        expected_app_name = "MyApp"
        expected_title = 'DjangoEsi'
        self.assertEqual(operation.future.request.headers["User-Agent"], f"{expected_app_name}/1.0.0 (email@example.com; +https://example.com) {expected_title}/1.0.0 (+https://gitlab.com/allianceauth/django-esi)")

@patch(MODULE_PATH + ".__title__", "Django-ESI")
@patch(MODULE_PATH + ".__version__", "1.0.0")
@requests_mock.Mocker()
class TestEsiClientProviderAppText(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        cls.spec = _load_json_file(SWAGGER_SPEC_PATH_MINIMAL)
        cls.status_response = {
            "players": 12345,
            "server_version": "1132976",
            "start_time": "2017-01-02T12:34:56Z",
        }

    @patch(MODULE_PATH + ".app_settings.ESI_USER_CONTACT_EMAIL", None)
    def test_defaults(self, requests_mocker) -> None:
        # This test is not expected to hit given that ESI_USER_CONTACT_EMAIL must be set
        # But here it is for completeness
        requests_mocker.register_uri("GET", url="https://esi.evetech.net/_latest/swagger.json", json=self.spec)
        requests_mocker.register_uri("GET", url="https://esi.evetech.net/latest/swagger.json", json=self.spec)
        requests_mocker.register_uri("GET", url="https://esi.evetech.net/v1/status/", json=self.status_response)
        client = EsiClientProvider().client
        # when
        operation = client.Status.get_status()
        # then
        expected_title = 'DjangoEsi'
        self.assertEqual(operation.future.request.headers["User-Agent"], f"{expected_title}/1.0.0 (None; +https://gitlab.com/allianceauth/django-esi)")

    @patch(MODULE_PATH + ".app_settings.ESI_USER_CONTACT_EMAIL", "email@example.com")
    def test_defaults_email(self, requests_mocker) -> None:
        # given
        requests_mocker.register_uri("GET", url="https://esi.evetech.net/_latest/swagger.json", json=self.spec)
        requests_mocker.register_uri("GET", url="https://esi.evetech.net/latest/swagger.json", json=self.spec)
        requests_mocker.register_uri("GET", url="https://esi.evetech.net/v1/status/", json=self.status_response)
        client = EsiClientProvider().client
        # when
        operation = client.Status.get_status()
        # then
        expected_title = 'DjangoEsi'
        self.assertEqual(operation.future.request.headers["User-Agent"], f"{expected_title}/1.0.0 (email@example.com; +https://gitlab.com/allianceauth/django-esi)")

    @patch(MODULE_PATH + ".app_settings.ESI_USER_CONTACT_EMAIL", None)
    def test_app_text(self, requests_mocker) -> None:
        # Deprecated
        # This test is not expected to hit given that ESI_USER_CONTACT_EMAIL must be set
        # given
        requests_mocker.register_uri("GET", url="https://esi.evetech.net/_latest/swagger.json", json=self.spec)
        requests_mocker.register_uri("GET", url="https://esi.evetech.net/latest/swagger.json", json=self.spec)
        requests_mocker.register_uri("GET", url="https://esi.evetech.net/v1/status/", json=self.status_response)
        client = EsiClientProvider(app_info_text="my-app v1.0.0").client
        # when
        operation = client.Status.get_status()
        # then
        expected_title = 'DjangoEsi'
        self.assertEqual(operation.future.request.headers["User-Agent"], f"my-app v1.0.0 (None) {expected_title}/1.0.0 (+https://gitlab.com/allianceauth/django-esi)",)

    @patch(MODULE_PATH + ".app_settings.ESI_USER_CONTACT_EMAIL", "email@example.com")
    def test_app_text_with_email(self, requests_mocker):
        # Deprecated
        # given
        requests_mocker.register_uri("GET", url="https://esi.evetech.net/_latest/swagger.json", json=self.spec)
        requests_mocker.register_uri("GET", url="https://esi.evetech.net/latest/swagger.json", json=self.spec)
        requests_mocker.register_uri("GET", url="https://esi.evetech.net/v1/status/", json=self.status_response)
        client = EsiClientProvider(app_info_text="my-app v1.0.0").client
        # when
        operation = client.Status.get_status()
        # then
        expected_title = 'DjangoEsi'
        self.assertEqual(operation.future.request.headers["User-Agent"], f"my-app v1.0.0 (email@example.com) {expected_title}/1.0.0 (+https://gitlab.com/allianceauth/django-esi)",)

    @patch(MODULE_PATH + ".app_settings.ESI_USER_CONTACT_EMAIL", "email@example.com")
    def test_ua_generator(self, requests_mocker):
        requests_mocker.register_uri("GET", url="https://esi.evetech.net/_latest/swagger.json", json=self.spec)
        requests_mocker.register_uri("GET", url="https://esi.evetech.net/latest/swagger.json", json=self.spec)
        requests_mocker.register_uri("GET", url="https://esi.evetech.net/v1/status/", json=self.status_response)
        client = EsiClientProvider(ua_appname="MyApp", ua_version="1.0.0").client
        # when
        operation = client.Status.get_status()
        # then
        expected_app_name = "MyApp"
        expected_title = 'DjangoEsi'
        self.assertEqual(operation.future.request.headers["User-Agent"], f"{expected_app_name}/1.0.0 (email@example.com) {expected_title}/1.0.0 (+https://gitlab.com/allianceauth/django-esi)")

    @patch(MODULE_PATH + ".app_settings.ESI_USER_CONTACT_EMAIL", "email@example.com")
    def test_ua_generator_with_url(self, requests_mocker):
        requests_mocker.register_uri("GET", url="https://esi.evetech.net/_latest/swagger.json", json=self.spec)
        requests_mocker.register_uri("GET", url="https://esi.evetech.net/latest/swagger.json", json=self.spec)
        requests_mocker.register_uri("GET", url="https://esi.evetech.net/v1/status/", json=self.status_response)
        client = EsiClientProvider(ua_appname="MyApp", ua_version="1.0.0", ua_url="https://example.com").client
        # when
        operation = client.Status.get_status()
        # then
        expected_app_name = "MyApp"
        expected_title = 'DjangoEsi'
        self.assertEqual(operation.future.request.headers["User-Agent"], f"{expected_app_name}/1.0.0 (email@example.com; +https://example.com) {expected_title}/1.0.0 (+https://gitlab.com/allianceauth/django-esi)")



from django.dispatch import Signal

"""
Basic signal sent after an ESI requests
    from django.dispatch import receiver
    from esi.signals import esi_request_statistics

    @receiver(esi_request_statistics)
    def esi_callback(sender, operation, status_code, headers, latency, bucket, **kwargs):
        # do stuff
        pass
"""
esi_request_statistics = Signal(
    # providing_args=[
    #     "operation",
    #     "status_code",
    #     "headers",
    #     "latency",
    #     "bucket"
    # ]
)



from datetime import timedelta
from unittest.mock import patch

from celery import current_app as celery_app
from django.utils.timezone import now

from esi.models import CallbackRedirect, Token
from esi.tasks import cleanup_callbackredirect, cleanup_token, cleanup_token_subset

from . import NoSocketsTestCase
from .factories_2 import TokenFactory, CallbackRedirectFactory

from math import ceil

MANAGERS_PATH = "esi.managers"
MODELS_PATH = "esi.models"
TASKS_PATH = "esi.tasks"


class CeleryTestCase(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        celery_app.conf.task_always_eager = True
        celery_app.conf.task_eager_propagates = True


class TestCleanupCallbackredirect(CeleryTestCase):
    def test_should_remove_expired(self) -> None:
        # given
        cb_valid = CallbackRedirectFactory()
        with patch("django.utils.timezone.now") as m:
            m.return_value = now() - timedelta(minutes=5, seconds=1)
            cb_expired = CallbackRedirectFactory()
        # when
        cleanup_callbackredirect.delay(max_age=300)
        # then
        self.assertTrue(CallbackRedirect.objects.filter(pk=cb_valid.pk).exists())
        self.assertFalse(CallbackRedirect.objects.filter(pk=cb_expired.pk).exists())


@patch(MANAGERS_PATH + '.app_settings.ESI_TOKEN_VALID_DURATION', 120)
@patch(MODELS_PATH + '.Token.refresh', spec=True)
class TestCleanupToken(CeleryTestCase):
    def test_should_delete_orphaned_tokens(self, mock_token_refresh) -> None:
        # given
        token_1 = TokenFactory(user=None)
        token_2 = TokenFactory()
        # when
        cleanup_token.delay()
        # then
        self.assertFalse(Token.objects.filter(pk=token_1.pk).exists())
        self.assertTrue(Token.objects.filter(pk=token_2.pk).exists())

    def test_should_refresh_expired_tokens_only(self, mock_token_refresh) -> None:
        # given
        TokenFactory()
        with patch("django.utils.timezone.now") as m:
            m.return_value = now() - timedelta(minutes=3)
            TokenFactory()
        # when
        cleanup_token.delay()
        # then
        self.assertEqual(mock_token_refresh.call_count, 1)


class TestCleanupTokenSubset(CeleryTestCase):
    @patch(MANAGERS_PATH + '.app_settings.ESI_TOKEN_VALID_DURATION', 120)
    @patch(TASKS_PATH + ".refresh_or_delete_token.apply_async")
    def test_should_delete_orphaned_tokens(self, mock_token_refresh) -> None:
        # given
        orphaned = TokenFactory(user=None)
        valid = TokenFactory()
        # when
        cleanup_token_subset.delay()
        # then
        self.assertFalse(Token.objects.filter(pk=orphaned.pk).exists())
        self.assertTrue(Token.objects.filter(pk=valid.pk).exists())

    @patch(MANAGERS_PATH + '.app_settings.ESI_TOKEN_VALID_DURATION', 1)
    @patch(TASKS_PATH + ".refresh_or_delete_token.apply_async")
    def test_should_refresh_fraction_of_expired_tokens(self, mock_token_refresh) -> None:
        # given
        num_expired = 100
        for _ in range(num_expired):
            TokenFactory(refresh_token="some_token")
        # patch time so all tokens are expired
        with patch("django.utils.timezone.now") as m:
            m.return_value = now() + timedelta(minutes=5)
            # when
            cleanup_token_subset.delay(fraction=10)
        # then
        expected_count = ceil(num_expired / 10)
        self.assertEqual(mock_token_refresh.call_count, expected_count)

    @patch(MANAGERS_PATH + '.app_settings.ESI_TOKEN_VALID_DURATION', 1)
    @patch("esi.tasks.refresh_or_delete_token.apply_async")
    def test_should_refresh_expired_tokens_only(self, mock_token_refresh) -> None:
        # given
        TokenFactory()
        with patch("django.utils.timezone.now") as m:
            m.return_value = now() - timedelta(minutes=3)
            TokenFactory()
        # when
        cleanup_token_subset.delay()
        # then
        self.assertEqual(mock_token_refresh.call_count, 1)

    @patch(MANAGERS_PATH + '.app_settings.ESI_TOKEN_VALID_DURATION', 1)
    @patch(TASKS_PATH + ".refresh_or_delete_token.apply_async")
    def test_should_log_and_exit_gracefully_with_no_tokens(self, mock_token_refresh) -> None:
        # when
        cleanup_token_subset.delay()
        # then
        self.assertEqual(Token.objects.count(), 0)
        mock_token_refresh.assert_not_called()



import logging
from datetime import timedelta
from math import ceil

from celery import shared_task

from django.utils import timezone

from .models import CallbackRedirect, Token

logger = logging.getLogger(__name__)


@shared_task
def cleanup_callbackredirect(max_age=300):
    """
    Delete old :model:`esi.CallbackRedirect` models.
    Accepts a max_age parameter, in seconds (default 300).
    """
    max_age = timezone.now() - timedelta(seconds=max_age)
    logger.debug(
        "Deleting all callback redirects created before %s",
        max_age.strftime("%b %d %Y %H:%M:%S")
    )
    CallbackRedirect.objects.filter(created__lte=max_age).delete()


@shared_task
def cleanup_token() -> None:
    """
    Delete Orphaned Tokens, then refresh or delete expired :model:`esi.Token` models.
    """
    orphaned_tokens = Token.objects.filter(user__isnull=True)
    if orphaned_tokens.exists():
        logger.info("Deleting %d orphaned tokens.", orphaned_tokens.count())
        orphaned_tokens.delete()

    expired_tokens = Token.objects.exclude(user__isnull=True).get_expired()
    if expired_tokens.exists():
        logger.info(
            "Triggering bulk refresh of %d expired tokens.", expired_tokens.count()
        )
        for token_pk in (
            expired_tokens.filter(refresh_token__isnull=False)
            .values_list("pk", flat=True)
        ):
            refresh_or_delete_token.apply_async(args=[token_pk], priority=8)


@shared_task
def cleanup_token_subset(fraction: int = 48) -> None:
    """
    Delete Orphaned Tokens, then refresh or delete a subset of expired :model:`esi.Token` models.

    This task operates on 1/fraction of the oldest tokens and can be called on a more regular schedule
    """
    orphaned_tokens = Token.objects.filter(user__isnull=True)
    if orphaned_tokens.exists():
        logger.info("Deleting %d orphaned tokens.", orphaned_tokens.count())
        orphaned_tokens.delete()

    expired_tokens = Token.objects.exclude(user__isnull=True).get_expired()
    expired_tokens_subset = expired_tokens.filter(
        refresh_token__isnull=False
    ).order_by("created")[:ceil(expired_tokens.count() / fraction)]

    if expired_tokens.exists():
        logger.info(
            f"Triggering bulk refresh of subset/possible {expired_tokens_subset.count()}/{expired_tokens.count()} expired tokens."
        )
        for token_pk in expired_tokens_subset.values_list("pk", flat=True):
            refresh_or_delete_token.apply_async(args=[token_pk], priority=8)


@shared_task
def refresh_or_delete_token(token_pk: int):
    token = Token.objects.get(pk=token_pk)
    token.refresh_or_delete()



# Generated by Django 3.2.12 on 2022-02-19 18:18

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('esi', '0010_set_new_tokens_to_sso_v2'),
    ]

    operations = [
        migrations.AlterField(
            model_name='token',
            name='character_id',
            field=models.IntegerField(db_index=True, help_text='The ID of the EVE character who authenticated by SSO.'),
        ),
        migrations.AlterField(
            model_name='token',
            name='character_name',
            field=models.CharField(db_index=True, help_text='The name of the EVE character who authenticated by SSO.', max_length=100),
        ),
        migrations.AlterField(
            model_name='token',
            name='character_owner_hash',
            field=models.CharField(db_index=True, help_text='The unique string identifying this character and its owning EVE account. Changes if the owning account changes.', max_length=254),
        ),
    ]



"""unit tests for esi checks"""

from django.conf import settings
from django.test import TestCase, override_settings

from ..checks import check_sso_application_settings


class TestCheckSsoApplicationSettings(TestCase):

    @override_settings(
        ESI_SSO_CLIENT_ID='123', ESI_SSO_CLIENT_SECRET='abc', ESI_SSO_CALLBACK_URL='xyz'
    )
    def test_settings_ok(self):
        errors = check_sso_application_settings()
        self.assertEqual(len(errors), 0)

    @override_settings(ESI_SSO_CLIENT_SECRET='abc', ESI_SSO_CALLBACK_URL='xyz')
    def test_settings_incomplete_ESI_SSO_CLIENT_ID(self):
        del settings.ESI_SSO_CLIENT_ID
        errors = check_sso_application_settings()
        self.assertEqual(len(errors), 1)

    @override_settings(ESI_SSO_CLIENT_ID='123', ESI_SSO_CALLBACK_URL='xyz')
    def test_settings_incomplete_ESI_SSO_CLIENT_SECRET(self):
        del settings.ESI_SSO_CLIENT_SECRET
        errors = check_sso_application_settings()
        self.assertEqual(len(errors), 1)

    @override_settings(ESI_SSO_CLIENT_ID='123', ESI_SSO_CLIENT_SECRET='abc')
    def test_settings_incomplete_ESI_SSO_CALLBACK_URL(self):
        del settings.ESI_SSO_CALLBACK_URL
        errors = check_sso_application_settings()
        self.assertEqual(len(errors), 1)

    @override_settings(ESI_SSO_CLIENT_ID='123', ESI_SSO_CLIENT_SECRET='abc', DEBUG=True)
    def test_settings_incomplete_debug_mode(self):
        del settings.ESI_SSO_CALLBACK_URL
        errors = check_sso_application_settings()
        self.assertEqual(len(errors), 1)

    @override_settings(
        ESI_SSO_CLIENT_ID='123', ESI_SSO_CLIENT_SECRET='abc', DEBUG=False
    )
    def test_settings_incomplete_non_debug_mode(self):
        del settings.ESI_SSO_CALLBACK_URL
        errors = check_sso_application_settings()
        self.assertEqual(len(errors), 1)



import logging
import time
from collections.abc import Callable
from functools import wraps
from typing import Any

from celery import Task

from django.core.cache import cache

from esi.exceptions import ESIBucketLimitException, ESIErrorLimitException
from esi.rate_limiting import ESIRateLimitBucket, ESIRateLimits

from .models import CallbackRedirect, Token

logger = logging.getLogger(__name__)


def _check_callback(request) -> Token | None:
    # ensure session installed in database
    if not request.session.exists(request.session.session_key):
        logger.debug("Creating new session for %s", request.user)
        request.session.create()

    # clean up callback redirect, pass token if new requested
    try:
        model = CallbackRedirect.objects.get(session_key=request.session.session_key)
        token = Token.objects.get(pk=model.token.pk)
        model.delete()
        logger.debug(
            "Retrieved new token from callback for %s session %s",
            request.user,
            request.session.session_key[:5])
        return token

    except (CallbackRedirect.DoesNotExist, Token.DoesNotExist, AttributeError):
        logger.debug(
            "No callback for %s session %s",
            request.user,
            request.session.session_key[:5],
            exc_info=True
        )
        return None


def tokens_required(scopes='', new=False):
    """
    Decorator for views to request an ESI Token.
    Accepts required scopes as a space-delimited string
    or list of strings of scope names.
    Can require a new token to be retrieved by SSO.
    Returns a QueryDict of Tokens.
    """

    def decorator(view_func):
        @wraps(view_func)
        def _wrapped_view(request, *args, **kwargs):

            # if we're coming back from SSO with a new token, return it
            token = _check_callback(request)
            if token:
                tokens = Token.objects.filter(pk=token.pk)
                logger.debug("Returning new token.")
                return view_func(request, tokens, *args, **kwargs)

            if not new:
                # ensure user logged in to check existing tokens
                if not request.user.is_authenticated:
                    logger.debug(
                        "Session %s is not logged in. Redirecting to login.",
                        request.session.session_key[:5]
                    )
                    from django.contrib.auth.views import redirect_to_login
                    return redirect_to_login(request.get_full_path())

                # collect tokens in db, check if still valid, return if any
                tokens = (
                    Token.objects
                    .filter(user__pk=request.user.pk)
                    .require_scopes(scopes)
                    .require_valid()
                )
                if tokens.exists():
                    logger.debug(
                        "Retrieved %s tokens for %s session %s",
                        tokens.count(),
                        request.user,
                        request.session.session_key[:5]
                    )
                    return view_func(request, tokens, *args, **kwargs)

            # trigger creation of new token via sso
            logger.debug(
                "No tokens identified for %s session %s. Redirecting to SSO.",
                request.user,
                request.session.session_key[:5]
            )
            from esi.views import sso_redirect
            return sso_redirect(request, scopes=scopes)

        return _wrapped_view

    return decorator


def token_required(scopes='', new=False):
    """
    Decorator for views which supplies a single,
    user-selected token for the view to process.
    Same parameters as tokens_required.
    """

    def decorator(view_func):
        @wraps(view_func)
        def _wrapped_view(request, *args, **kwargs):

            # if we're coming back from SSO with a new token, return it
            token = _check_callback(request)
            if token:
                logger.debug(
                    "Got new token from %s session %s. Returning to view.",
                    request.user,
                    request.session.session_key[:5]
                )
                return view_func(request, token, *args, **kwargs)

            # if we're selecting a token, return it
            if request.method == 'POST':
                if request.POST.get("_add", False):
                    logger.debug(
                        "%s has selected to add new token. Redirecting to SSO.",
                        request.user
                    )
                    # user has selected to add a new token
                    from esi.views import sso_redirect
                    return sso_redirect(request, scopes=scopes)

                token_pk = request.POST.get('_token', None)
                if token_pk:
                    logger.debug(
                        "%s has selected token %s", request.user, token_pk
                    )
                    try:
                        token = Token.objects.get(pk=token_pk)
                        # ensure token belongs to this user and has required scopes
                        if (
                            (
                                (token.user and token.user == request.user)
                                or not token.user
                            )
                            and Token.objects
                            .filter(pk=token_pk)
                            .require_scopes(scopes)
                            .require_valid()
                            .exists()
                        ):
                            logger.debug(
                                "Selected token fulfills requirements of view. "
                                "Returning."
                            )
                            return view_func(request, token, *args, **kwargs)

                    except Token.DoesNotExist:
                        logger.debug("Token %s not found.", token_pk)

            if not new:
                # present the user with token choices
                tokens = (
                    Token.objects
                    .filter(user__pk=request.user.pk)
                    .require_scopes(scopes)
                    .require_valid()
                )
                if tokens.exists():
                    logger.debug(
                        "Returning list of available tokens for %s.", request.user
                    )
                    from esi.views import select_token
                    return select_token(request, scopes=scopes, new=new)
                else:
                    logger.debug(
                        "No tokens found for %s session %s with scopes %s",
                        request.user,
                        request.session.session_key[:5],
                        scopes
                    )

            # prompt the user to add a new token
            logger.debug(
                "Redirecting %s session %s to SSO.",
                request.user,
                request.session.session_key[:5]
            )
            from esi.views import sso_redirect
            return sso_redirect(request, scopes=scopes)

        return _wrapped_view

    return decorator


def single_use_token(scopes='', new=False):
    """
    Decorator for views which supplies a single use token granted via sso login
    regardless of login state.
    Same parameters as tokens_required.
    """

    def decorator(view_func):
        @wraps(view_func)
        def _wrapped_view(request, *args, **kwargs):

            # if we're coming back from SSO for a new token, return it
            token = _check_callback(request)
            if token:
                logger.debug(
                    "Got new token from session %s. Returning to view.",
                    request.session.session_key[:5]
                )
                return view_func(request, token, *args, **kwargs)

            # prompt the user to login for a new token
            logger.debug(
                "Redirecting session %s to SSO.", request.session.session_key[:5]
            )
            from esi.views import sso_redirect
            return sso_redirect(request, scopes=scopes)

        return _wrapped_view

    return decorator


def wait_for_esi_errorlimit_reset(cache_key="esi_error_limit_reset", poll_interval=1) -> Callable[..., Callable[..., Any]]:
    """
    Decorator to apply a polling sleep while the ESI Server/Client is in an Error Limit state
    The preferred non-blocking method is to retry your tasks after the limit reset time has passed

    Args:
        cache_key (str, optional):  NOT USUALLY CHANGED. Defaults to "esi_error_limit_reset".
        poll_interval (int, optional): Interval in seconds to poll redis. Defaults to 1.
    """
    def decorator(func):
        def wrapper(*args, **kwargs):
            reset = cache.get(cache_key)
            if reset is not None:
                logger.error(f"ESI Error Limited, waiting {reset}s before retrying...")
                while cache.get(cache_key):
                    time.sleep(poll_interval)
            return func(*args, **kwargs)
        return wrapper
    return decorator


def esi_rate_limiter_bucketed(bucket: ESIRateLimitBucket, raise_on_limit: bool = True):
    """
    Decorator for custom manual rate limits on some endpoints to apply a polling sleep while the bucket is exhausted.
    MARKET_DATA_HISTORY
    CHARACTER_CORPORATION_HISTORY
    The preferred non-blocking method is to retry your tasks after the limit reset time has passed


    Args:
        bucket (ESIRateLimitBucket): The Bucket to rate limit against
        raise_on_limit (bool, optional): Whether to raise an Exception when the limit is reached. Defaults to True.
    """
    # TODO Investigate esi cache hits.
    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            ESIRateLimits.check_decr_bucket(bucket, raise_on_limit)
            return func(*args, **kwargs)
        return wrapper
    return decorator


def rate_limit_retry_task(func):
    """
    Retry celery task on ESI rate limit exhaustion.
    """
    @wraps(func)
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except (ESIBucketLimitException, ESIErrorLimitException) as ex:
            context: Task = args[0]
            if isinstance(context, Task):
                # Cool down for a window duration.
                # If we start tracking requests on our end we can math this better.
                context.retry(countdown=ex.reset)
            else:
                logger.error(
                    f"{ex.bucket} Exhausted. Unable to retry Task first arg is not a Task instance."
                )
                raise ex
    return wrapper



from django.core.management.base import BaseCommand

class Command(BaseCommand):
    help = 'Clear ESI OpenAPI specification document caches.'

    def handle(self, *args, **options):
        self.stdout.write("Finding all ESI Spec Caches.")
        try:
            from django_redis import get_redis_connection
            _client = get_redis_connection("default")
        except (NotImplementedError, ModuleNotFoundError):
            from django.core.cache import caches
            default_cache = caches['default']
            _client = default_cache.get_master_client()

        keys = _client.keys(":?:ESI_API_CACHE_*")
        self.stdout.write(f"Found {len(keys)} cached entries")

        if keys:
            deleted = _client.delete(*keys)
            self.stdout.write(f"Deleted {deleted} entries")



import logging
import pathlib
import warnings
import datetime as dt
from hashlib import md5
from timeit import default_timer
from typing import Any

from aiopenapi3 import OpenAPI, FileSystemLoader
from aiopenapi3._types import ResponseDataType, ResponseHeadersType
from aiopenapi3.errors import HTTPServerError as base_HTTPServerError
from aiopenapi3.errors import HTTPClientError as base_HTTPClientError
from aiopenapi3.request import OperationIndex, RequestBase
from esi.aiopenapi3.client import SpecCachingClient
from esi.rate_limiting import ESIRateLimitBucket, ESIRateLimits, interval_to_seconds
from httpx import (
    AsyncClient, Client, HTTPStatusError, RequestError, Response, Timeout,
)
from tenacity import (
    AsyncRetrying, Retrying, retry_if_exception, stop_after_attempt,
    wait_combine, wait_exponential,
)

from django.core.cache import cache
from django.utils.text import slugify

from esi import app_settings
from esi.exceptions import HTTPClientError, HTTPServerError, HTTPNotModified
from esi.aiopenapi3.plugins import (
    Add304ContentType, DjangoESIInit, PatchCompatibilityDatePlugin,
    Trim204ContentType, MinifySpec
)
from esi.exceptions import ESIErrorLimitException
from esi.models import Token
from esi.signals import esi_request_statistics
from esi.stubs import ESIClientStub

from . import __title__, __url__, __version__
from .helpers import pascal_case_string

logger = logging.getLogger(__name__)

ETAG_EXPIRY = 60 * 60 * 24 * 7  # 7 days


def _time_to_expiry(expires_header: str) -> int:
    """Calculate cache TTL from Expires header
    Args:
        expires_header (str): The value of the Expires header '%a, %d %b %Y %H:%M:%S %Z'
    Returns:
        int: The cache TTL in seconds
    """
    try:
        expires_dt = dt.datetime.strptime(str(expires_header), '%a, %d %b %Y %H:%M:%S %Z')
        if expires_dt.tzinfo is None:
            expires_dt = expires_dt.replace(tzinfo=dt.timezone.utc)
        return max(int((expires_dt - dt.datetime.now(dt.timezone.utc)).total_seconds()), 0)
    except ValueError:
        return 0


def _httpx_exceptions(exc: BaseException) -> bool:
    """
    Helper function for HTTP Retries, what various exceptions and status codes should we retry on.
    ESI has some weird behaviours
    """
    if isinstance(exc, ESIErrorLimitException):
        return False
    if isinstance(exc, RequestError):
        return True
    if isinstance(exc, HTTPStatusError) and getattr(exc.response, "status_code", None) in {502, 503, 504}:
        return True
    return False


def http_retry_sync() -> Retrying:
    return Retrying(
        retry=retry_if_exception(_httpx_exceptions),
        wait=wait_combine(
            wait_exponential(multiplier=1, min=1, max=10),
        ),
        stop=stop_after_attempt(3),
        reraise=True,
    )


async def http_retry_async() -> AsyncRetrying:  # pragma: no cover
    return AsyncRetrying(
        retry=retry_if_exception(_httpx_exceptions),
        wait=wait_combine(
            wait_exponential(multiplier=1, min=1, max=10),
        ),
        stop=stop_after_attempt(3),
        reraise=True,
    )


def _load_plugins(app_name, tags: list[str] = [], operations: list[str] = []):
    """Load the plugins to make ESI work with this lib.

    Args:
        app_name (str): app name to use for internal etags
    """
    return [
        PatchCompatibilityDatePlugin(),
        Trim204ContentType(),
        Add304ContentType(),
        DjangoESIInit(app_name),
        MinifySpec(tags, operations)
    ]


def _load_aiopenapi_client_sync(
        spec_url: str,
        compatibility_date: str,
        app_name: str,
        user_agent: str,
        tenant: str,
        spec_file: str | None = None,
        tags: list[str] = [],
        operations: list[str] = []) -> OpenAPI:
    """Create an OpenAPI3 Client from Spec

    Args:
        spec_url (str): _description_
        compatibility_date (str): _description_
        app_name (str): _description_
        user_agent (str): _description_
        tenant (str): _description_
        spec_file (str | None, optional): _description_. Defaults to None.

    Returns:
        OpenAPI: aiopenapi3 Client Class
    """
    headers = {
        "User-Agent": user_agent,
        "X-Tenant": tenant,
        "X-Compatibility-Date": compatibility_date
    }

    def session_factory(**kwargs) -> Client:
        kwargs.pop("headers", None)
        return SpecCachingClient(
            headers=headers,
            timeout=Timeout(
                connect=app_settings.ESI_REQUESTS_CONNECT_TIMEOUT,
                read=app_settings.ESI_REQUESTS_READ_TIMEOUT,
                write=app_settings.ESI_REQUESTS_WRITE_TIMEOUT,
                pool=app_settings.ESI_REQUESTS_POOL_TIMEOUT
            ),
            http2=True,
            **kwargs
        )
    if spec_file:
        return OpenAPI.load_file(
            url=spec_url,
            path=spec_file,
            session_factory=session_factory,
            loader=FileSystemLoader(pathlib.Path(spec_file)),
            use_operation_tags=True,
            plugins=_load_plugins(app_name, tags, operations)
        )
    else:
        return OpenAPI.load_sync(
            url=spec_url,
            session_factory=session_factory,
            use_operation_tags=True,
            plugins=_load_plugins(app_name, tags, operations)
        )


async def _load_aiopenapi_client_async(
        spec_url: str,
        compatibility_date: str,
        app_name: str,
        user_agent: str,
        tenant: str,
        spec_file: str | None = None) -> OpenAPI:  # pragma: no cover
    """Create an OpenAPI3 Client from Spec Async

    Args:
        spec_url (str): _description_
        compatibility_date (str): _description_
        user_agent (str): _description_
        tenant (str): _description_
        spec_file (str | None, optional): _description_. Defaults to None.

    Returns:
        OpenAPI: aiopenapi3 Client Class
    """
    headers = {
        "User-Agent": user_agent,
        "X-Tenant": tenant,
        "X-Compatibility-Date": compatibility_date
    }

    def session_factory(**kwargs) -> AsyncClient:
        kwargs.pop("headers", None)
        return AsyncClient(
            headers=headers,
            timeout=Timeout(
                connect=app_settings.ESI_REQUESTS_CONNECT_TIMEOUT,
                read=app_settings.ESI_REQUESTS_READ_TIMEOUT,
                write=app_settings.ESI_REQUESTS_WRITE_TIMEOUT,
                pool=app_settings.ESI_REQUESTS_POOL_TIMEOUT
            ),
            http2=True,
            **kwargs
        )
    if spec_file:
        # TODO find a async way to load from file?
        return OpenAPI.load_file(
            url=spec_url,
            path=spec_file,
            session_factory=session_factory,
            use_operation_tags=True,
            plugins=_load_plugins(app_name)
        )
    else:
        return await OpenAPI.load_async(
            url=spec_url,
            session_factory=session_factory,
            use_operation_tags=True,
            plugins=_load_plugins(app_name)
        )


def _build_user_agent(ua_appname: str, ua_version: str, ua_url: str | None = None) -> str:
    """
    AppName/1.2.3 (foo@example.com; +https://gitlab.com/) Django-ESI/1.2.3 (+https://gitlab.com/allianceauth/django-esi)
    Contact Email will be inserted from app_settings.
    Args:
        ua_appname (str): Application Name, PascalCase
        ua_version (str): Application Version, SemVer
        ua_url (str | None): Application URL (Optional)
    Returns:
        str: User-Agent string
    """

    # Enforce PascalCase for `ua_appname` and strip whitespace
    sanitized_ua_appname = pascal_case_string(ua_appname)
    sanitized_appname = pascal_case_string(__title__)

    return (
        f"{sanitized_ua_appname}/{ua_version} "
        f"({app_settings.ESI_USER_CONTACT_EMAIL}{f'; +{ua_url})' if ua_url else ')'} "
        f"{sanitized_appname}/{__version__} (+{__url__})"
    )


def _get_spec_url() -> str:
    return f"{app_settings.ESI_API_URL}meta/openapi.json"


def esi_client_factory_sync(
        compatibility_date: str,
        ua_appname: str, ua_version: str, ua_url: str | None = None,
        spec_file: str | None = None,
        tenant: str = "tranquility",
        tags: list[str] = [], operations: list[str] = [],
        **kwargs) -> OpenAPI:
    """Generate a new OpenAPI ESI client.
    Args:
        compatibility_date (str): "YYYY-MM-DD" The latest version of ESI your client is tested with
        ua_appname (str): Application Name, PascalCase
        ua_version (str): Application Version, SemVer
        ua_url (str, optional): Application URL (Optional). Defaults to None.
        spec_file (str | None, optional): Specification file path (Optional). Defaults to None.
        tenant (str, optional): Tenant ID (Optional). Defaults to "tranquility".
    Returns:
        OpenAPI: OpenAPI ESI Client
    """
    user_agent = _build_user_agent(ua_appname, ua_version, ua_url)
    spec_url = _get_spec_url()
    return _load_aiopenapi_client_sync(
        spec_url,
        compatibility_date,
        ua_appname,
        user_agent,
        tenant,
        spec_file,
        tags,
        operations
    )


async def esi_client_factory_async(
        compatibility_date: str,
        ua_appname: str, ua_version: str, ua_url: str | None = None,
        spec_file: str | None = None,
        tenant: str = "tranquility",
        **kwargs) -> OpenAPI:  # pragma: no cover
    """Generate a new OpenAPI ESI client.
    Args:
        compatibility_date (str): "YYYY-MM-DD" The latest version of ESI your client is tested with
        ua_appname (str): Application Name, PascalCase
        ua_version (str): Application Version, SemVer
        ua_url (str | None, optional): Application URL (Optional). Defaults to None.
        spec_file (str | None, optional): Specification file path (Optional). Defaults to None.
        tenant (str, optional): Tenant ID (Optional). Defaults to "tranquility".
    Returns:
        OpenAPI: OpenAPI ESI Client
    """
    user_agent = _build_user_agent(ua_appname, ua_version, ua_url)
    spec_url = _get_spec_url()
    return await _load_aiopenapi_client_async(spec_url, compatibility_date, ua_appname, user_agent, tenant, spec_file)


class BaseEsiOperation():
    def __init__(self, operation, api: OpenAPI) -> None:
        self.method, self.url, self.operation, self.extra = operation
        self.api = api
        self.token: Token | None = None
        self.bucket: ESIRateLimitBucket | None = None

        self._args = []
        self._kwargs = {}

        self._set_bucket()

    def __call__(self, *args, **kwargs) -> "BaseEsiOperation":
        self._args = args
        self._kwargs = kwargs
        return self

    def _unnormalize_parameters(self, params: dict[str, Any]) -> dict[str, Any]:
        """UN-Normalize Pythonic parameter names back to OpenAPI names.

        Converts pythonic keys like "Accept_Language" to "Accept-Language" when/if
        a non pythonic (usually) hyphenated form exists in the operation's parameter list. Performs
        case-insensitive matching and only rewrites when there's a known
        parameter with hyphens, leaving normal snake_case params (e.g.
        "type_id") untouched.
        Args:
            params: Raw parameters collected from the call
        Returns:
            dict: Parameters with keys aligned to the OpenAPI spec
        """
        try:
            spec_param_names = [p.name for p in getattr(self.operation, "parameters", [])]
        except Exception:
            spec_param_names = []

        # Exact and case-insensitive lookup maps
        spec_param_set = set(spec_param_names)
        spec_param_map_ci = {n.lower(): n for n in spec_param_names}

        normalized: dict[str, Any] = {}
        for k, v in params.items():
            # Fast path: exact match
            if k in spec_param_set:
                normalized[k] = v
                continue

            # Try hyphen variant
            k_dash = k.replace("_", "-")
            if k_dash in spec_param_set:
                normalized[k_dash] = v
                continue

            # Case-insensitive fallbacks
            kl = k.lower()
            if kl in spec_param_map_ci:
                normalized[spec_param_map_ci[kl]] = v
                continue

            k_dash_l = k_dash.lower()
            if k_dash_l in spec_param_map_ci:
                normalized[spec_param_map_ci[k_dash_l]] = v
                continue

            # Unknown to the spec; pass through as-is (aiopenapi3 will validate)
            normalized[k] = v

        return normalized

    def _etag_key(self) -> str:
        """Generate a key name used to cache etag responses based on app_name and cache_key
        Returns:
            str: Key
        """
        # ignore the token this will break the cache
        return f"{slugify(self.api.app_name)}_etag_{self._cache_key()}"  # type: ignore app_name is added by a plugin

    def _cache_key(self) -> str:
        """Generate a key name used to cache responses based on method, url, args, kwargs
        Returns:
            str: Key
        """
        # ignore the token this will break the cache
        ignore_keys = [
            "token",
        ]
        _kwargs = {key: value for key, value in self._kwargs.items() if key not in ignore_keys}
        data = (self.method + self.url + str(self._args) + str(_kwargs)).encode('utf-8')
        str_hash = md5(data).hexdigest()  # nosec B303
        return f'esi_{str_hash}'

    def _extract_body_param(self) -> Token | None:
        """Pop the request body from parameters to be able to check the param validity
        Returns:
            Any | None: the request body
        """
        _body = self._kwargs.pop("body", None)
        if _body and not getattr(self.operation, "requestBody", False):
            raise ValueError("Request Body provided on endpoint with no request body parameter.")
        return _body

    def _extract_token_param(self) -> Token | None:
        """Pop token from parameters or use the Client wide token if set
        Returns:
            Token | None: The token to use for the request
        """
        _token = self._kwargs.pop("token", None)
        if _token and not getattr(self.operation, "security", False):
            raise ValueError("Token provided on public endpoint")
        return self.token or _token

    def _has_page_param(self) -> bool:
        """Check if this operation supports Offset Based Pagination.
        Returns:
            bool: True if page parameters are present, False otherwise
        """
        return any(p.name == "page" for p in self.operation.parameters)

    def _has_cursor_param(self) -> bool:
        """Check if this operation supports Cursor Based Pagination.
        Returns:
            bool: True if cursor parameters are present, False otherwise
        """
        return any(p.name == "before" or p.name == "after" for p in self.operation.parameters)

    def _get_cache(self, cache_key: str, etag: str | None) -> tuple[ResponseHeadersType | None, Any, Response | None]:
        """Retrieve cached response and validate expiry
        Args:
            cache_key (str): The cache key to retrieve
        Returns:
            tuple[ResponseHeadersType | None, Any, Response | None]: The cached response,
            or None if not found or expired
        """
        if not app_settings.ESI_CACHE_RESPONSE:
            return None, None, None

        try:
            cached_response = cache.get(cache_key)
        except Exception as e:
            logger.error(f"Cache retrieve failed {e}", exc_info=True)
            return None, None, None

        if cached_response:
            logger.debug(f"Cache Hit {self.url}")
            expiry = _time_to_expiry(str(cached_response.headers.get('Expires')))

            # force check to ensure cache isn't expired
            if expiry < 0:
                logger.warning("Cache expired by %d seconds, forcing expiry", expiry)
                return None, None, None

            # check if etag is same before building models from cache
            if etag:
                if cached_response.headers.get('ETag') == etag:
                    # refresh/store the etag's TTL
                    self._send_signal(
                        status_code=0,  # this is a cached response less a 304
                        headers=cached_response.headers,
                        latency=0
                    )
                    self._store_etag(cached_response.headers)
                    raise HTTPNotModified(
                        status_code=304,
                        headers=cached_response.headers
                    )

            # build models
            headers, data = self.parse_cached_request(cached_response)
            return headers, data, cached_response

        return None, None, None

    def _store_etag(self, headers: dict):
        """
            Store response etag in cache for 7 days
        """
        if "ETag" in headers:
            cache.set(self._etag_key(), headers["ETag"], timeout=ETAG_EXPIRY)

    def _update_etag_ttl(self):
        """
            reset etag ttl in cache.
        """
        cache.expire(self._etag_key(), timeout=ETAG_EXPIRY)

    def _clear_etag(self):
        """ Delete the cached etag for this operation.
        """
        try:
            cache.delete(self._etag_key())
        except Exception as e:
            logger.error(f"Failed to delete etag {e}", exc_info=True)

    def _store_cache(self, cache_key: str, response) -> None:
        """ Store the response in cache for expiry TTL.
        Args:
            cache_key (str): The cache key to store the response under
            response (Response): The response object to cache
        """
        if not app_settings.ESI_CACHE_RESPONSE:
            return

        expires = response.headers.get("Expires")
        ttl = _time_to_expiry(expires) if expires else 0
        if ttl > 0:
            try:
                cache.set(cache_key, response, ttl)
            except Exception as e:
                logger.error(f"Failed to cache {e}", exc_info=True)

    def _clear_cache(self):
        """ Delete the cached data for this operation.
        """
        try:
            cache.delete(self._cache_key())
        except Exception as e:
            logger.error(f"Failed to delete cache {e}", exc_info=True)

    def _validate_token_scopes(self, token: Token) -> None:
        """Validate that the token provided has the required scopes for this ESI operation.
        """
        token_scopes = set(token.scopes.all().values_list("name", flat=True))
        try:
            required_scopes = set(getattr(getattr(self.operation, "security", [])[0], "root", {}).get("OAuth2", []))
        except KeyError:
            required_scopes = []
        missing_scopes = [x for x in required_scopes if x not in token_scopes]
        if len(missing_scopes) > 0:
            raise ValueError(f"Token Missing Scopes - {missing_scopes}")

    def parse_cached_request(self, cached_response) -> tuple[ResponseHeadersType, ResponseDataType]:
        req = self.api.createRequest(
            f"{self.operation.tags[0]}.{self.operation.operationId}"
        )
        return req._process_request(cached_response)

    def _set_bucket(self):
        """Setup the rate bucket"""
        _rate_limit = getattr(self.operation, "extensions", {}).get("rate-limit", False)
        if _rate_limit:
            _key = _rate_limit["group"]
            if self.token:
                _key = f"{_key}:{self.token.character_id}"
            self.bucket = ESIRateLimitBucket(
                _key,
                _rate_limit["max-tokens"],
                interval_to_seconds(_rate_limit["window-size"])
            )

    def _send_signal(self, status_code: int, headers: dict = {}, latency: float = 0) -> None:
        """
            Dispatch the esi request statistics signal
        """
        esi_request_statistics.send(
            sender=self.__class__,
            operation=self.operation.operationId,
            status_code=status_code,
            headers=headers,
            latency=latency,
            bucket=self.bucket.slug if self.bucket else ""
        )


class EsiOperation(BaseEsiOperation):
    def __skip__process__headers__(
        self, result, headers: dict[str, str], expected_response
    ):
        """Return all headers always"""
        return headers

    def _make_request(
            self,
            parameters: dict[str, Any],
            etag: str | None = None) -> RequestBase.Response:

        reset = cache.get("esi_error_limit_reset")
        if reset is not None:
            # Hard exception here if there is still an open Error Limit
            # developers need to either decorators.wait_for_esi_error_limit_reset()
            # or handle this by pushing their celery tasks back
            raise ESIErrorLimitException(reset=reset)

        if self.bucket:
            """Check Rate Limit"""
            ESIRateLimits.check_bucket(self.bucket)

        retry = http_retry_sync()

        def __func():
            req = self.api.createRequest(f"{self.operation.tags[0]}.{self.operation.operationId}")

            # We want all headers from ESI
            # don't check/parse them against the spec and return them all
            # TODO Investigate if this is a bug with aiopenapi or a spec compliance issue
            req._process__headers = self.__skip__process__headers__

            if self.token:
                self.api.authenticate(OAuth2=True)  # make the lib happy
                if isinstance(self.token, str):
                    # Fallback older Django-ESI Behaviour
                    # Deprecated
                    req.req.headers["Authorization"] = f"Bearer {self.token}"
                    warnings.warn(
                        "Passing an Access Token string directly is deprecated."
                        "Doing so will Skip Validation of Scopes"
                        "Please use a Token object instead.",
                        DeprecationWarning,
                        stacklevel=2
                    )
                else:
                    self._validate_token_scopes(self.token)
                    req.req.headers["Authorization"] = f"Bearer {self.token.valid_access_token()}"
            if etag:
                req.req.headers["If-None-Match"] = etag
            _response = req.request(data=self.body, parameters=self._unnormalize_parameters(parameters))

            if self.bucket and "x-ratelimit-remaining" in _response.result.headers:
                logger.debug(
                    "ESI Rate-Limit: "
                    f"'{_response.result.headers.get('x-ratelimit-group')}' - "
                    f"Used {_response.result.headers.get('x-ratelimit-used')} - "
                    f"{_response.result.headers.get('x-ratelimit-remaining')} / "
                    f"({_response.result.headers.get('x-ratelimit-limit')})"
                )
                ESIRateLimits.set_bucket(
                    self.bucket,
                    _response.result.headers.get("x-ratelimit-remaining")
                )

            return _response
        return retry(__func)

    def result(
            self,
            use_etag: bool = True,
            return_response: bool = False,
            force_refresh: bool = False,
            use_cache: bool = True,
            **extra) -> tuple[Any, Response] | Any:
        """Executes the request and returns the response from ESI for the current operation.
        Raises:
            ESIErrorLimitException: _description_
            ESIBucketLimitException: _description_
        Returns:
            _type_: _description_
        """
        _t = default_timer()
        self.token = self._extract_token_param()

        if self.token:
            self._set_bucket()

        self.body = self._extract_body_param()
        parameters = self._kwargs | extra
        cache_key = self._cache_key()
        etag_key = self._etag_key()
        etag = None

        if force_refresh:
            self._clear_cache()
            self._clear_etag()

        if use_etag:
            etag = cache.get(etag_key)

        headers, data, response = self._get_cache(cache_key, etag=etag) if use_cache else (None, None, None)

        if response and use_cache:
            expiry = _time_to_expiry(str(headers.get('Expires')))
            if expiry < 0:
                logger.warning(
                    "cache expired by %d seconds, Forcing expiry", expiry
                )
                response = None
                headers = None
                data = None

        if not response:
            logger.debug(f"Cache Miss {self.url}")
            try:
                headers, data, response = self._make_request(parameters, etag)
            # Shim our exceptions into Django-ESI
            except base_HTTPServerError as e:
                self._send_signal(
                    status_code=e.status_code,
                    headers=e.headers,
                    latency=default_timer() - _t
                )
                raise HTTPServerError(
                    status_code=e.status_code,
                    headers=e.headers,
                    data=e.data
                )

            except base_HTTPClientError as e:
                self._send_signal(
                    status_code=e.status_code,
                    headers=e.headers,
                    latency=default_timer() - _t
                )

                if e.status_code == 420:
                    reset = e.headers.get("X-RateLimit-Reset", None)
                    if reset:
                        reset = int(reset)
                        cache.set("esi_error_limit_reset", reset, timeout=reset)
                    raise ESIErrorLimitException(reset=reset)

                raise HTTPClientError(
                    status_code=e.status_code,
                    headers=e.headers,
                    data=e.data
                )

            self._send_signal(
                status_code=response.status_code,
                headers=response.headers,
                latency=default_timer() - _t
            )

            # store the ETAG in cache
            self._store_etag(response.headers)

            # Throw a 304 exception for catching.
            if response.status_code == 304:
                # refresh/store the etag's TTL
                self._update_etag_ttl()
                raise HTTPNotModified(
                    status_code=304,
                    headers=response.headers
                )

            # last step store cache after 304 logic, we dont want to catch the 304 `None` responses
            self._store_cache(cache_key, response)

        else:
            # send signal for cached data too
            self._send_signal(
                status_code=0,
                headers=response.headers,
                latency=default_timer() - _t
            )

        return (data, response) if return_response else data

    def results(
            self,
            use_etag: bool = True,
            return_response: bool = False,
            force_refresh: bool = False,
            use_cache: bool = True,
            **extra) -> tuple[list[Any], Response | Any | None] | list[Any]:
        all_results: list[Any] = []
        last_response: Response | None = None
        """Executes the request and returns the response from ESI for the current
        operation. Response will include all pages if there are more available.

        Returns:
            _type_: _description_
        """
        if self._has_page_param():
            current_page = 1
            total_pages = 1
            count_pages_etag_hit = 0
            force_refetch = False
            last_headers = None
            while current_page <= total_pages:
                self._kwargs["page"] = current_page
                try:
                    data, response = self.result(
                        # use cache where we can, but ignore etags if we are re-fetching
                        use_etag=use_etag and not force_refetch,
                        return_response=True,
                        force_refresh=force_refresh,
                        use_cache=use_cache,
                        **extra
                    )
                    last_response = response
                    all_results.extend(data if isinstance(data, list) else [data])
                    total_pages = int(response.headers.get("X-Pages", 1))
                    logger.debug(
                        f"ESI Page Fetched {self.url} - {current_page}/{total_pages}"
                    )
                    current_page += 1

                except HTTPNotModified as e:
                    count_pages_etag_hit += 1
                    total_pages = int(e.headers.get("X-Pages", 1))
                    logger.debug(
                        f"ESI Page hit ETAG {self.url} - {current_page}/{total_pages}"
                    )
                    current_page += 1
                    last_headers = e.headers

                if (
                    current_page > total_pages and
                    count_pages_etag_hit != total_pages and
                    count_pages_etag_hit > 0
                ):
                    # Not all pages hit etag, so refetch all
                    force_refetch = True
                    current_page = 1
                    count_pages_etag_hit = 0
                    all_results = []
                    last_response = None

                elif (
                    current_page > total_pages and
                    count_pages_etag_hit == total_pages
                ):
                    # All etags hit raise 304
                    raise HTTPNotModified(
                        status_code=304,
                        headers=last_headers
                    )

        elif self._has_cursor_param():
            # Untested, there are no cursor based endpoints in ESI
            params = self._kwargs.copy()
            params.update(extra)
            for cursor_param in ("after", "before"):
                if params.get(cursor_param):
                    break
            else:
                cursor_param = "after"
            while True:
                data, response = self.result(
                    use_etag=use_etag,
                    return_response=True,
                    force_refresh=force_refresh,
                    use_cache=use_cache,
                    **params
                )
                last_response = response
                if not data:
                    break
                all_results.extend(data if isinstance(data, list) else [data])
                cursor_token = {k.lower(): v for k, v in response.headers.items()}.get(cursor_param)
                if not cursor_token:
                    break
                params[cursor_param] = cursor_token

        else:
            data, response = self.result(
                use_etag=use_etag,
                return_response=True,
                force_refresh=force_refresh,
                use_cache=use_cache,
                **extra
            )
            all_results.extend(data if isinstance(data, list) else [data])
            last_response = response

        return (all_results, last_response) if return_response else all_results

    def results_localized(
            self,
            languages: list[str] | str | None = None,
            **kwargs
    ) -> dict[str, list[Any]]:
        """Executes the request and returns the response from ESI for all default languages and pages (if any).
        Args:
            languages: (list[str], str, optional) language(s) to return instead of default languages
        Raises:
            ValueError: Invalid or Not Supported Language Code ...
        Returns:
            dict[str, list[Any]]: Dict of all responses with the language code as keys.
        """
        if not languages:
            my_languages = list(app_settings.ESI_LANGUAGES)
        else:
            my_languages = []
            for lang in dict.fromkeys(languages):
                if lang not in app_settings.ESI_LANGUAGES:
                    raise ValueError('Invalid or Not Supported Language Code: %s' % lang)
                my_languages.append(lang)

        return {
            language: self.results(accept_language=language, **kwargs)
            for language in my_languages
        }

    def required_scopes(self) -> list[str]:
        """Return a simple list of scopes required for an endpoint. #Requires loading and processing a client
        Returns:
            list[str]: List of Scopes Required
        """
        try:
            if not getattr(self.operation, "security", False):
                return []  # No Scopes Required
            else:
                return list(getattr(getattr(self.operation, "security", [])[0], "root", {}).get("OAuth2", []))
        except (IndexError, KeyError):
            return []


class EsiOperationAsync(BaseEsiOperation):  # pragma: no cover
    async def _make_request(
            self,
            parameters: dict[str, Any],
            etag: str | None = None
    ) -> RequestBase.Response:

        reset = cache.get("esi_error_limit_reset")
        if reset is not None:
            # Hard exception here if there is still an open rate limit
            # developers need to either decorators.wait_for_esi_error_limit_reset()
            # or handle this by pushing their celery tasks back
            raise ESIErrorLimitException(reset=reset)

        async for attempt in http_retry_async():
            with attempt:
                req = self.api.createRequest(f"{self.operation.tags[0]}.{self.operation.operationId}")
                if self.token:
                    self.api.authenticate(OAuth2=True)  # make the lib happy
                    self._validate_token_scopes(self.token)
                    req.req.headers["Authorization"] = f"Bearer {self.token.valid_access_token()}"
                if etag:
                    req.req.headers["If-None-Match"] = etag
                return req.request(parameters=self._unnormalize_parameters(parameters))
        # Should never be reached because AsyncRetrying always yields at least once
        raise RuntimeError("Retry loop exited without performing a request")

    async def result(
        self,
        etag: str | None = None,
        return_response: bool = False,
        use_cache: bool = True,
        **extra
    ) -> tuple[Any, Response] | Any:
        self.token = self._extract_token_param()
        parameters = self._kwargs | extra
        cache_key = self._cache_key()
        etag_key = f"{cache_key}_etag"

        if not etag and app_settings.ESI_CACHE_RESPONSE:
            etag = cache.get(etag_key)

        headers, data, response = self._get_cache(cache_key, etag)

        if response and use_cache:
            expiry = _time_to_expiry(str(headers.get('Expires')))
            if expiry < 0:
                logger.warning(
                    "cache expired by %d seconds, Forcing expiry", expiry
                )
                response = None
                headers = None
                data = None

        if not response:
            logger.debug(f"Cache Miss {self.url}")
            try:
                headers, data, response = await self._make_request(parameters, etag)
                if response.status_code == 420:
                    reset = response.headers.get("X-RateLimit-Reset", None)
                    cache.set("esi_error_limit_reset", reset, timeout=reset)
                    raise ESIErrorLimitException(reset=reset)
                self._store_cache(cache_key, response)
                self._store_etag(response.headers)
            # Shim our exceptions into Django-ESI
            except base_HTTPServerError as e:
                raise HTTPServerError(
                    status_code=e.status_code,
                    headers=e.headers,
                    data=e.data
                )

            except base_HTTPClientError as e:
                raise HTTPClientError(
                    status_code=e.status_code,
                    headers=e.headers,
                    data=e.data
                )

            # Throw a 304 exception for catching.
            if response.status_code == 304:
                # refresh/store the etag's TTL
                self._store_etag(response.headers)
                raise HTTPNotModified(
                    status_code=304,
                    headers=response.headers
                )

        return (data, response) if return_response else data

    async def results(
            self,
            etag: str | None = None,
            return_response: bool = False,
            use_cache: bool = True,
            **extra
    ) -> tuple[list[Any], Response | Any | None] | list[Any]:
        all_results = []
        last_response = None

        if self._has_page_param():
            current_page = 1
            total_pages = 1
            while current_page <= total_pages:
                self._kwargs["page"] = current_page
                data, response = await self.result(etag=etag, return_response=True, use_cache=use_cache, **extra)
                last_response = response
                all_results.extend(data if isinstance(data, list) else [data])
                total_pages = int(response.headers.get("X-Pages", 1))
                logger.debug(
                    f"ESI Page Fetched {self.url} - {current_page}/{total_pages}"
                )
                current_page += 1
        # elif self._has_cursor_param():
            # TODO
        else:
            data, response = await self.result(etag=etag, return_response=True, use_cache=use_cache, **extra)
            all_results.extend(data if isinstance(data, list) else [data])
            last_response = response

        return (all_results, last_response) if return_response else all_results

    def results_localized(
            self,
            languages: list[str] | str | None = None,
            **extra) -> dict[str, list[Any]]:
        """Executes the request and returns the response from ESI for all default languages and pages (if any).
        Args:
            languages: (list[str], str, optional) language(s) to return instead of default languages
        Raises:
            ValueError: Invalid or Not Supported Language Code ...
        Returns:
            dict[str, list[Any]]: Dict of all responses with the language code as keys.
        """
        if not languages:
            my_languages = list(app_settings.ESI_LANGUAGES)
        else:
            my_languages = []
            for lang in dict.fromkeys(languages):
                if lang not in app_settings.ESI_LANGUAGES:
                    raise ValueError('Invalid or Not Supported Language Code: %s' % lang)
                my_languages.append(lang)

        return {
            language: self.results(accept_language=language, **extra)
            for language in my_languages
        }

    def required_scopes(self) -> list[str]:
        """Return a simple list of scopes required for an endpoint. #Requires loading and processing a client
        Returns:
            list[str]: List of Scopes Required
        """
        try:
            if not getattr(self.operation, "security", False):
                return []  # No Scopes Required
            else:
                return list(getattr(getattr(self.operation, "security", [])[0], "root", {}).get("OAuth2", []))
        except (IndexError, KeyError):
            return []


class ESITag:
    """
    API Tag Wrapper, providing access to Operations within a tag
    Assets, Characters, etc.
    """

    def __init__(self, operation, api) -> None:
        self._oi = operation._oi
        self._operations = operation._operations
        self.api = api

    def __getattr__(self, name: str) -> EsiOperation:
        if name not in self._operations:
            raise AttributeError(
                f"Operation '{name}' not found in tag '{self._oi}'. "
                f"Available operations: {', '.join(sorted(self._operations.keys()))}"
            )
        return EsiOperation(self._operations[name], self.api)


class ESITagAsync():  # pragma: no cover
    """
    Async API Tag Wrapper, providing access to Operations within a tag
    Assets, Characters, etc.
    """

    def __init__(self, operation, api) -> None:
        self._oi = operation._oi
        self._operations = operation._operations
        self.api = api

    def __getattr__(self, name: str) -> EsiOperationAsync:
        if name not in self._operations:
            raise AttributeError(
                f"Operation '{name}' not found in tag '{self._oi}'. "
                f"Available operations: {', '.join(sorted(self._operations.keys()))}"
            )
        return EsiOperationAsync(self._operations[name], self.api)


class ESIClient(ESIClientStub):
    """
    Base ESI Client, provides access to Tags Assets, Characters, etc.
    or Raw aiopenapi3 via sad smiley ._.
    """

    def __init__(self, api: OpenAPI) -> None:
        self.api = api
        self._tags = set(api._operationindex._tags.keys())

    def __getattr__(self, tag: str) -> ESITag | OperationIndex:
        # underscore returns the raw aiopenapi3 client
        if tag == "_":
            return self.api._operationindex

        # convert pythonic Planetary_Interaction to Planetary Interaction
        if "_" in tag:
            tag = tag.replace("_", " ")

        if tag in set(self.api._operationindex._tags.keys()):
            return ESITag(self.api._operationindex._tags[tag], self.api)

        raise AttributeError(
            f"Tag '{tag}' not found. "
            f"Available tags: {', '.join(sorted(self._tags))}"
        )

    def purge_all_etags(self):
        """ Delete all stored etags from the cache for this application

        TODO: consider making this more config agnostic
        """
        try:
            # new lib
            from django_redis import get_redis_connection
            _client = get_redis_connection("default")
        except (NotImplementedError, ModuleNotFoundError):
            # old lib
            from django.core.cache import caches
            default_cache = caches['default']
            _client = default_cache.get_master_client()  # type: ignore

        keys = _client.keys(f":?:{slugify(self.api.app_name)}_etag_*")  # type: ignore app_name is added by a plugin
        if keys:
            deleted = _client.delete(*keys)

        logger.info(f"Deleted {deleted} etag keys")

        return deleted


class ESIClientAsync(ESIClientStub):  # pragma: no cover
    """
    Async Base ESI Client, provides access to Tags Assets, Characters, etc.
    or Raw aiopenapi3 via sad smiley ._.
    """

    def __init__(self, api: OpenAPI) -> None:
        self.api = api
        self._tags = set(api._operationindex._tags.keys())

    def __getattr__(self, tag: str) -> ESITagAsync | OperationIndex:
        # underscore returns the raw aiopenapi3 client
        if tag == "_":
            return self.api._operationindex

        # convert pythonic Planetary_Interaction to Planetary Interaction
        if "_" in tag:
            tag = tag.replace("_", " ")

        if tag in set(self.api._operationindex._tags.keys()):
            return ESITagAsync(self.api._operationindex._tags[tag], self.api)

        raise AttributeError(
            f"Tag '{tag}' not found. "
            f"Available tags: {', '.join(sorted(self._tags))}"
        )


class ESIClientProvider:
    """Class for providing a single ESI client instance for a whole app
    * Note that one of either `tags` or `operations` must be provided to reduce memory footprint of the client
    * When `DEBUG=False`, not supplying either will raise an AttributeError.
    Args:
        compatibility_date (str | date): The compatibility date for the ESI client.
        ua_appname (str): Name of the App for generating a User-Agent,
        ua_version (str): Version of the App for generating a User-Agent,
        ua_url (str, Optional): URL To the Source Code or Documentation for generating a User-Agent,
        spec_file (str, Optional): Absolute path to a OpenApi 3.1 spec file to load.
        tenant (str, Optional): The ESI tenant to use (default: "tranquility").
        operations (list[str], Optional*): List of operations to filter the spec down.
        tags (list[str], Optional*): List of tags to filter the spec down.
    Functions:
        client(): ESIClient
        client_async(): ESIClientAsync
    """

    _client: ESIClient | None = None
    _client_async: ESIClientAsync | None = None

    def __init__(
        self,
        compatibility_date: str | dt.date,
        ua_appname: str,
        ua_version: str,
        ua_url: str | None = None,
        spec_file: None | str = None,
        tenant: str = "tranquility",
        operations: list[str] = [],
        tags: list[str] = [],
        **kwargs
    ) -> None:
        if type(compatibility_date) is dt.date:
            self._compatibility_date: str = self._date_to_string(compatibility_date)
        else:
            self._compatibility_date: str = str(compatibility_date)
        self._ua_appname = ua_appname
        self._ua_version = ua_version
        self._ua_url = ua_url
        self._spec_file = spec_file
        self._tenant = tenant
        self._kwargs = kwargs
        self._operations = operations
        self._tags = tags

    @property
    def client(self) -> ESIClient:
        if self._client is None:
            api = esi_client_factory_sync(
                compatibility_date=self._compatibility_date,
                ua_appname=self._ua_appname,
                ua_version=self._ua_version,
                ua_url=self._ua_url,
                spec_file=self._spec_file,
                tenant=self._tenant,
                operations=self._operations,
                tags=self._tags,
                **self._kwargs)
            self._client = ESIClient(api)
        return self._client

    @property
    async def client_async(self) -> ESIClientAsync:  # pragma: no cover
        if self._client_async is None:
            api = await esi_client_factory_async(
                compatibility_date=self._compatibility_date,
                ua_appname=self._ua_appname,
                ua_version=self._ua_version,
                ua_url=self._ua_url,
                spec_file=self._spec_file,
                tenant=self._tenant,
                operations=self._operations,
                tags=self._tags,
                **self._kwargs)
            self._client_async = ESIClientAsync(api)
        return self._client_async

    @classmethod
    def _date_to_string(cls, compatibility_date: dt.date) -> str:
        """Turns a date object in a compatibility_date string"""
        return f"{compatibility_date.year}-{compatibility_date.month:02}-{compatibility_date.day:02}"

    def __str__(self) -> str:
        return f"ESIClientProvider - {self._ua_appname} ({self._ua_version})"



from django.contrib import admin
from django.contrib.auth import get_user_model
from django.db.models import QuerySet

from .models import CallbackRedirect, Scope, Token

admin.site.register(CallbackRedirect)


@admin.register(Scope)
class ScopeAdmin(admin.ModelAdmin):
    list_display = ('name', 'help_text')


@admin.register(Token)
class TokenAdmin(admin.ModelAdmin):
    def get_queryset(self, request) -> QuerySet["Token"]:
        qs = super().get_queryset(request)
        return qs.select_related('user').prefetch_related('scopes')

    @admin.display(
        description='Scopes'
    )
    def get_scopes(self, obj) -> str:
        return ", ".join([x.name for x in obj.scopes.all()])

    User = get_user_model()
    list_display = ('user', 'character_name', 'get_scopes')
    search_fields = ['user__%s' % User.USERNAME_FIELD, 'character_name', 'scopes__name']
    list_filter = (
        ('user', admin.RelatedOnlyFieldListFilter),
        'character_name'
    )
    ordering = ('user',)
    readonly_fields = (
        'character_id',
        'character_name',
        'token_type',
        'character_owner_hash',
        'scopes',
        'sso_version'
    )



from unittest.mock import patch
from io import StringIO

from django.contrib.auth.models import User
from django.test import TestCase
from django.core.management import call_command as base_call_command

from . import _generate_token, _store_as_Token

from esi.errors import (
    TokenInvalidError,
    NotRefreshableTokenError
)

from oauthlib.oauth2.rfc6749.errors import InvalidGrantError, \
    InvalidTokenError, \
    InvalidClientIdError


@patch('esi.models.Token.delete')
@patch('requests_oauthlib.OAuth2Session.refresh_token')
@patch('esi.models.Token.refresh')
class TestSSOMigrations(TestCase):
    """tests for SSOv1 to SSOv2 Migration command"""

    def setUp(self):

        character_id = 1001
        character_name = 'Batman'

        self.user = User.objects.create_user(
            character_name,
            'abc@example.com',
            'password'
        )
        self.token_v1 = _store_as_Token(
            _generate_token(
                character_id=character_id,
                character_name=character_name,
                scopes=['esi-universe.read_structures.v1'],
                sso_version=1
            ),
            self.user
        )
        self.token_v3 = _store_as_Token(
            _generate_token(
                character_id=character_id,
                character_name=character_name,
                scopes=['esi-universe.read_structures.v1'],
                sso_version=1
            ),
            self.user
        )
        self.token_v4 = _store_as_Token(
            _generate_token(
                character_id=character_id,
                character_name=character_name,
                scopes=['esi-universe.read_structures.v1'],
                sso_version=1
            ),
            self.user
        )
        self.token_v5 = _store_as_Token(
            _generate_token(
                character_id=character_id,
                character_name=character_name,
                scopes=['esi-universe.read_structures.v1'],
                sso_version=1
            ),
            self.user
        )
        self.token_v2 = _store_as_Token(
            _generate_token(
                character_id=character_id,
                character_name=character_name,
                scopes=['publicdata']

            ),
            self.user
        )

    def call_command(self, *args, **kwargs):
        std_out = StringIO()
        std_err = StringIO()
        base_call_command(
            "migrate_to_ssov2",
            *args,
            stdout=std_out,
            stderr=std_err,
            **kwargs,
        )
        return (std_out.getvalue(), std_err.getvalue())

    def test_ignore_sso_v1(
        self, mock_v2_refresh, mock_v1_refresh, mock_delete
    ):
        mock_v1_refresh.return_value = {
            'access_token': 'access_token_2',
            'refresh_token': 'refresh_token_2'
        }

        self.call_command("--skip-v1-checks", "-n 1")

        self.assertEqual(mock_v2_refresh.call_count, 1)
        self.assertEqual(mock_v1_refresh.call_count, 0)
        self.assertEqual(mock_delete.call_count, 0)

    def test_normal_refresh(
        self, mock_v2_refresh, mock_v1_refresh, mock_delete
    ):

        mock_v1_refresh.return_value = {
            'access_token': 'access_token_2',
            'refresh_token': 'refresh_token_2'
        }
        self.call_command("-n 1")

        self.assertEqual(mock_v2_refresh.call_count, 1)
        self.assertEqual(mock_v1_refresh.call_count, 1)
        self.assertEqual(mock_delete.call_count, 0)

    def test_normal_refresh_all(
        self, mock_v2_refresh, mock_v1_refresh, mock_delete
    ):

        mock_v1_refresh.return_value = {
            'access_token': 'access_token_2',
            'refresh_token': 'refresh_token_2'
        }
        self.call_command()

        self.assertEqual(mock_v2_refresh.call_count, 4)
        self.assertEqual(mock_v1_refresh.call_count, 4)
        self.assertEqual(mock_delete.call_count, 0)

    def test_fail_v1_pre_refresh_ignore(
        self, mock_v2_refresh, mock_v1_refresh, mock_delete
    ):

        mock_v1_refresh.side_effect = InvalidGrantError
        self.call_command("-n 1")

        self.assertEqual(mock_v2_refresh.call_count, 0)
        self.assertEqual(mock_v1_refresh.call_count, 1)
        self.assertEqual(mock_delete.call_count, 0)

    def test_fail_v1_pre_refresh_delete_InvalidGrantError(
        self, mock_v2_refresh, mock_v1_refresh, mock_delete
    ):

        mock_v1_refresh.side_effect = InvalidGrantError
        self.call_command("--purge", "-n 1")

        self.assertEqual(mock_v2_refresh.call_count, 0)
        self.assertEqual(mock_v1_refresh.call_count, 1)
        self.assertEqual(mock_delete.call_count, 1)

    def test_fail_v1_pre_refresh_delete_InvalidTokenError(
        self, mock_v2_refresh, mock_v1_refresh, mock_delete
    ):

        mock_v1_refresh.side_effect = InvalidTokenError
        self.call_command("--purge", "-n 1")

        self.assertEqual(mock_v2_refresh.call_count, 0)
        self.assertEqual(mock_v1_refresh.call_count, 1)
        self.assertEqual(mock_delete.call_count, 1)

    def test_fail_v1_pre_refresh_delete_InvalidClientIdError(
        self, mock_v2_refresh, mock_v1_refresh, mock_delete
    ):

        mock_v1_refresh.side_effect = InvalidClientIdError
        self.call_command("--purge", "-n 1")

        self.assertEqual(mock_v2_refresh.call_count, 0)
        self.assertEqual(mock_v1_refresh.call_count, 1)
        self.assertEqual(mock_delete.call_count, 1)

    def test_fail_v1_pre_refresh_delete_Exception(
        self, mock_v2_refresh, mock_v1_refresh, mock_delete
    ):

        mock_v1_refresh.side_effect = Exception
        self.call_command("--purge", "-n 1")

        self.assertEqual(mock_v2_refresh.call_count, 0)
        self.assertEqual(mock_v1_refresh.call_count, 1)
        self.assertEqual(mock_delete.call_count, 1)

    def test_fail_v2_refresh_ignore(
        self, mock_v2_refresh, mock_v1_refresh, mock_delete
    ):

        mock_v2_refresh.side_effect = TokenInvalidError
        self.call_command("--purge", "--skip-v1-checks", "-n 1")

        self.assertEqual(mock_v2_refresh.call_count, 1)
        self.assertEqual(mock_v1_refresh.call_count, 0)
        self.assertEqual(mock_delete.call_count, 1)

    def test_fail_v2_refresh_delete(
        self, mock_v2_refresh, mock_v1_refresh, mock_delete
    ):

        mock_v2_refresh.side_effect = TokenInvalidError
        self.call_command("--purge", "--skip-v1-checks", "-n 1")

        self.assertEqual(mock_v2_refresh.call_count, 1)
        self.assertEqual(mock_v1_refresh.call_count, 0)
        self.assertEqual(mock_delete.call_count, 1)

    def test_v1_pass_v2_fail_v1_pass(
        self, mock_v2_refresh, mock_v1_refresh, mock_delete
    ):

        mock_v1_refresh.return_value = {
            'access_token': 'access_token_2',
            'refresh_token': 'refresh_token_2'
        }
        mock_v2_refresh.side_effect = TokenInvalidError

        self.call_command("--purge", "-n 1")

        self.assertEqual(mock_v2_refresh.call_count, 1)
        self.assertEqual(mock_v1_refresh.call_count, 2)
        self.assertEqual(mock_delete.call_count, 0)

    def test_v1_pass_v2_fail_v1_fail_ignore(
        self, mock_v2_refresh, mock_v1_refresh, mock_delete
    ):

        mock_v1_refresh.side_effect = [{
            'access_token': 'access_token_2',
            'refresh_token': 'refresh_token_2'
        }, InvalidGrantError]
        mock_v2_refresh.side_effect = TokenInvalidError

        self.call_command("-n 1")

        self.assertEqual(mock_v2_refresh.call_count, 1)
        self.assertEqual(mock_v1_refresh.call_count, 2)
        self.assertEqual(mock_delete.call_count, 0)

    def test_v1_pass_v2_fail_v1_fail_delete(
        self, mock_v2_refresh, mock_v1_refresh, mock_delete
    ):

        mock_v1_refresh.side_effect = [{
            'access_token': 'access_token_2',
            'refresh_token': 'refresh_token_2'
        }, InvalidGrantError]
        mock_v2_refresh.side_effect = TokenInvalidError

        self.call_command("--purge", "-n 1")

        self.assertEqual(mock_v2_refresh.call_count, 1)
        self.assertEqual(mock_v1_refresh.call_count, 2)
        self.assertEqual(mock_delete.call_count, 1)

    def test_v2_fail_non_refresh_delete(
        self, mock_v2_refresh, mock_v1_refresh, mock_delete
    ):

        mock_v1_refresh.side_effect = [{
            'access_token': 'access_token_2',
            'refresh_token': 'refresh_token_2'
        }]
        mock_v2_refresh.side_effect = NotRefreshableTokenError

        self.call_command("--purge", "-n 1")

        self.assertEqual(mock_v2_refresh.call_count, 1)
        self.assertEqual(mock_v1_refresh.call_count, 1)
        self.assertEqual(mock_delete.call_count, 1)

    def test_v2_fail_ignore(
        self, mock_v2_refresh, mock_v1_refresh, mock_delete
    ):

        mock_v1_refresh.side_effect = [{
            'access_token': 'access_token_2',
            'refresh_token': 'refresh_token_2'
        }]
        mock_v2_refresh.side_effect = TokenInvalidError

        self.call_command("--skip-v1-checks", "-n 1")

        self.assertEqual(mock_v2_refresh.call_count, 1)
        self.assertEqual(mock_v1_refresh.call_count, 0)
        self.assertEqual(mock_delete.call_count, 0)

    def test_v2_fail_non_refresh_ignore(
        self, mock_v2_refresh, mock_v1_refresh, mock_delete
    ):

        mock_v1_refresh.side_effect = [{
            'access_token': 'access_token_2',
            'refresh_token': 'refresh_token_2'
        }]
        mock_v2_refresh.side_effect = NotRefreshableTokenError

        self.call_command("-n 1")

        self.assertEqual(mock_v2_refresh.call_count, 1)
        self.assertEqual(mock_v1_refresh.call_count, 1)
        self.assertEqual(mock_delete.call_count, 0)



# flake8: noqa
"""script for testing connection pool size with live requests to ESI

Run this script directly. Make sure to also set the environment variable
DJANGO_PROJECT_PATH and DJANGO_SETTINGS_MODULE to match your setup:

You can see the result in your main log file of your Django installation.
If your connection pool size is too small you will see the following warning in your
log file: "Connection pool is full, discarding connection: esi.evetech.net"
To pass this test successfully ESI_CONNECTION_POOL_MAXSIZE must equal to number
of MAX_WORKER (e.g. 20)

Example:
export DJANGO_PROJECT_PATH="/home/erik997/dev/python/django-dev/mysite"
export DJANGO_SETTINGS_MODULE="mysite.settings"

"""

# start django project
import os
import sys

if not 'DJANGO_PROJECT_PATH' in os.environ:
    print('DJANGO_PROJECT_PATH is not set')
    exit(1)

if not 'DJANGO_SETTINGS_MODULE' in os.environ:
    print('DJANGO_SETTINGS_MODULE is not set')
    exit(1)

sys.path.insert(0, os.environ['DJANGO_PROJECT_PATH'])
import django
django.setup()

# normal imports
import concurrent.futures
import logging

from django.core.cache import cache
from esi.clients import EsiClientProvider

MAX_WORKER = 20

cache.clear()
esi = EsiClientProvider()

logger = logging.getLogger('__name__')
logger.level = logging.DEBUG
logger.propagate = True


def thread_func(type_id):
    logger.info('Fetching type with ID %d from ESI', type_id)
    esi.client.Universe.get_universe_types_type_id(type_id=type_id).result()

def main():
    print('Script started...')
    entity_ids = esi.client.Universe.get_universe_types().results()[1000:2000]
    logger.info('Start fetching %d types' , len(entity_ids))

    with concurrent.futures.ThreadPoolExecutor(max_workers=MAX_WORKER) as executor:
        executor.map(thread_func, entity_ids)

    logger.info('Finished fetching %d types', len(entity_ids))
    print('DONE')


if __name__ == '__main__':
    main()
